#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov 29, 2018
@author: Quentin Lutz <qlutz@enst.fr>
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.utils import Bunch


def house(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """House graph.

    * Undirected graph
    * 5 nodes, 6 edges

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    adjacency or graph : Union[sparse.csr_matrix, Bunch]
        Adjacency matrix or graph with metadata (positions).

    Example
    -------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> adjacency.shape
    (5, 5)

    """
    row = np.array([0, 0, 1, 1, 2, 3])
    col = np.array([1, 4, 2, 4, 3, 4])
    adjacency = sparse.csr_matrix((np.ones(len(row), dtype=int), (row, col)), shape=(5, 5))
    adjacency = (adjacency + adjacency.T).astype(bool)

    if metadata:
        x = np.array([0, -1, -1, 1, 1])
        y = np.array([2, 1, -1, -1, 1])
        graph = Bunch()
        graph.adjacency = adjacency
        graph.position = np.vstack((x, y)).T
        graph.name = 'house'
        return graph
    else:
        return adjacency


def bow_tie(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Bow tie graph.

    * Undirected graph
    * 5 nodes, 6 edges

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    adjacency or graph : Union[sparse.csr_matrix, Bunch]
        Adjacency matrix or graph with metadata (positions).

    Example
    -------
    >>> from sknetwork.data import bow_tie
    >>> adjacency = bow_tie()
    >>> adjacency.shape
    (5, 5)
    """
    row = np.array([0, 0, 0, 0, 1, 3])
    col = np.array([1, 2, 3, 4, 2, 4])
    adjacency = sparse.csr_matrix((np.ones(len(row), dtype=int), (row, col)), shape=(5, 5))
    adjacency = (adjacency + adjacency.T).astype(bool)

    if metadata:
        x = np.array([0, -1, 1, -1, 1])
        y = np.array([0, 1, 1, -1, -1])
        graph = Bunch()
        graph.adjacency = adjacency
        graph.position = np.vstack((x, y)).T
        graph.name = 'bow_tie'
        return graph
    else:
        return adjacency


def karate_club(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Karate club graph.

    * Undirected graph
    * 34 nodes, 78 edges
    * 2 labels

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    adjacency or graph : Union[sparse.csr_matrix, Bunch]
        Adjacency matrix or graph with metadata (labels, positions).

    Example
    -------
    >>> from sknetwork.data import karate_club
    >>> adjacency = karate_club()
    >>> adjacency.shape
    (34, 34)

    References
    ----------
    Zachary's karate club graph
    https://en.wikipedia.org/wiki/Zachary%27s_karate_club
    """
    row = np.array(
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
         1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3,
         3, 4, 4, 5, 5, 5, 6, 8, 8, 8, 9, 13, 14, 14, 15, 15, 18,
         18, 19, 20, 20, 22, 22, 23, 23, 23, 23, 23, 24, 24, 24, 25, 26, 26,
         27, 28, 28, 29, 29, 30, 30, 31, 31, 32])
    col = np.array(
        [1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 13, 17, 19, 21, 31, 2,
         3, 7, 13, 17, 19, 21, 30, 3, 7, 8, 9, 13, 27, 28, 32, 7, 12,
         13, 6, 10, 6, 10, 16, 16, 30, 32, 33, 33, 33, 32, 33, 32, 33, 32,
         33, 33, 32, 33, 32, 33, 25, 27, 29, 32, 33, 25, 27, 31, 31, 29, 33,
         33, 31, 33, 32, 33, 32, 33, 32, 33, 33])
    adjacency = sparse.csr_matrix((np.ones(len(row), dtype=bool), (row, col)), shape=(34, 34))
    adjacency = sparse.csr_matrix(adjacency + adjacency.T, dtype=bool)

    if metadata:
        labels = np.array(
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        x = np.array(
            [0.04,  0.24,  0.01,  0.13,  0.02, -0.08,  0.04,  0.21,  0.08, -0.11, -0.13, -0.28,  0.2,  0.08,
             0.23,  0.06, -0.06,  0.32, 0.15,  0.19,  0.27,  0.39, -0.04, -0.26, -0.51, -0.49, -0.19, -0.28,
             -0.11, -0.17,  0.22, -0.21,  0.03, 0])
        y = np.array(
            [-0.33, -0.15, -0.01, -0.28, -0.64, -0.75, -0.76, -0.25,  0.09, 0.23, -0.62, -0.4, -0.53, -0.07,
             0.55,  0.64, -1., -0.42, 0.6, -0.01,  0.45, -0.34,  0.61,  0.41,  0.14,  0.28,  0.68, 0.21,
             0.12,  0.54,  0.19,  0.09,  0.38,  0.33])
        graph = Bunch()
        graph.adjacency = adjacency
        graph.labels = labels
        graph.position = np.vstack((x, y)).T
        graph.name = 'karate_club'
        return graph
    else:
        return adjacency


def miserables(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Co-occurrence graph of the characters in the novel Les miserables by Victor Hugo.

    * Undirected graph
    * 77 nodes, 508 edges
    * Names of characters

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    adjacency or graph : Union[sparse.csr_matrix, Bunch]
        Adjacency matrix or graph with metadata (names, positions).

    Example
    -------
    >>> from sknetwork.data import miserables
    >>> adjacency = miserables()
    >>> adjacency.shape
    (77, 77)
    """
    row = np.array(
        [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  2,  2,  3, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
         11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12,
         16, 16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 17, 17, 18, 18, 18, 18, 18, 19, 19, 19, 19,
         20, 20, 20, 21, 21, 22, 23, 23, 23, 23, 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 24, 25, 25, 25,
         25, 25, 25, 25, 25, 25, 25, 25, 25, 25, 26, 26, 26, 26, 26, 26, 26, 27, 27, 27, 27, 27, 27, 27,
         27, 27, 27, 27, 27, 28, 28, 29, 29, 29, 29, 29, 30, 34, 34, 34, 34, 35, 35, 35, 36, 36, 37, 39,
         39, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 47, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48,
         48, 48, 48, 48, 48, 48, 49, 49, 49, 49, 49, 51, 51, 51, 51, 54, 55, 55, 55, 55, 55, 55, 55, 55,
         55, 57, 57, 57, 57, 57, 57, 57, 57, 58, 58, 58, 58, 58, 58, 58, 58, 58, 58, 59, 59, 59, 59, 59,
         59, 59, 60, 60, 60, 60, 60, 60, 61, 61, 61, 61, 61, 62, 62, 62, 62, 62, 63, 63, 63, 63, 64, 64,
         64, 65, 65, 66, 68, 68, 68, 68, 69, 69, 69, 70, 70, 71, 73])
    col = np.array(
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 3, 11, 11, 11, 12, 13, 14,
         15, 23, 24, 25, 26, 27, 28, 29, 31, 32, 33, 34, 35, 36, 37, 38, 43,
         44, 48, 49, 51, 55, 58, 64, 68, 69, 70, 71, 72, 23, 17, 18, 19, 20,
         21, 22, 23, 26, 55, 18, 19, 20, 21, 22, 23, 19, 20, 21, 22, 23, 20,
         21, 22, 23, 21, 22, 23, 22, 23, 23, 24, 25, 27, 29, 30, 31, 25, 26,
         27, 41, 42, 50, 68, 69, 70, 26, 27, 39, 40, 41, 42, 48, 55, 68, 69,
         70, 71, 75, 27, 43, 49, 51, 54, 55, 72, 28, 29, 31, 33, 43, 48, 58,
         68, 69, 70, 71, 72, 44, 45, 34, 35, 36, 37, 38, 31, 35, 36, 37, 38,
         36, 37, 38, 37, 38, 38, 52, 55, 42, 55, 57, 62, 68, 69, 70, 71, 75,
         47, 48, 55, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 68, 69, 71, 73,
         74, 75, 76, 50, 51, 54, 55, 56, 52, 53, 54, 55, 55, 56, 57, 58, 59,
         61, 62, 63, 64, 65, 58, 59, 61, 62, 63, 64, 65, 67, 59, 60, 61, 62,
         63, 64, 65, 66, 70, 76, 60, 61, 62, 63, 64, 65, 66, 61, 62, 63, 64,
         65, 66, 62, 63, 64, 65, 66, 63, 64, 65, 66, 76, 64, 65, 66, 76, 65,
         66, 76, 66, 76, 76, 69, 70, 71, 75, 70, 71, 75, 71, 75, 75, 74])
    data = np.array(
        [1, 8, 10, 1, 1, 1, 1, 2, 1, 5, 6, 3, 3, 1, 1, 1, 1,
         1, 9, 7, 12, 31, 17, 8, 2, 3, 1, 2, 3, 3, 2, 2, 2, 3,
         1, 1, 2, 2, 19, 4, 1, 1, 1, 1, 1, 1, 2, 4, 4, 4, 3,
         3, 3, 3, 1, 1, 4, 4, 3, 3, 3, 3, 4, 3, 3, 3, 3, 4,
         3, 3, 3, 5, 4, 4, 4, 4, 4, 2, 1, 5, 1, 1, 2, 13, 4,
         1, 2, 1, 1, 1, 1, 1, 1, 5, 1, 1, 3, 2, 1, 2, 5, 6,
         4, 1, 3, 1, 1, 3, 2, 1, 21, 2, 1, 1, 1, 1, 1, 1, 6,
         1, 2, 1, 1, 1, 3, 2, 2, 2, 1, 1, 1, 2, 3, 2, 2, 2,
         2, 2, 2, 2, 2, 2, 1, 1, 2, 5, 1, 1, 1, 1, 1, 1, 1,
         1, 2, 4, 1, 7, 6, 1, 2, 7, 5, 5, 3, 1, 1, 1, 1, 2,
         2, 1, 1, 1, 9, 1, 12, 1, 1, 1, 2, 6, 1, 1, 1, 7, 5,
         1, 9, 1, 5, 2, 1, 2, 1, 2, 2, 1, 1, 3, 15, 4, 6, 17,
         4, 10, 5, 3, 1, 1, 2, 5, 13, 5, 9, 5, 1, 2, 3, 2, 2,
         2, 1, 6, 3, 6, 5, 1, 6, 12, 5, 2, 1, 4, 5, 1, 1, 7,
         3, 1, 2, 1, 1, 6, 4, 2, 3, 4, 2, 3, 2, 1, 1, 3])
    adjacency = sparse.csr_matrix((data, (row, col)), shape=(77, 77))
    adjacency = adjacency + adjacency.T

    if metadata:
        names = ['Myriel', 'Napoleon', 'Mlle Baptistine', 'Mme Magloire', 'Countess de Lo', 'Geborand',
                 'Champtercier', 'Cravatte', 'Count', 'Old man', 'Labarre', 'Valjean', 'Marguerite', 'Mme Der',
                 'Isabeau', 'Gervais', 'Tholomyes', 'Listolier', 'Fameuil', 'Blacheville', 'Favourite', 'Dahlia',
                 'Zephine', 'Fantine', 'Mme Thenardier', 'Thenardier', 'Cosette', 'Javert', 'Fauchelevent',
                 'Bamatabois', 'Perpetue', 'Simplice', 'Scaufflaire', 'Woman1', 'Judge', 'Champmathieu', 'Brevet',
                 'Chenildieu', 'Cochepaille', 'Pontmercy', 'Boulatruelle', 'Eponine', 'Anzelma', 'Woman2',
                 'Mother Innocent', 'Gribier', 'Jondrette', 'Mme Burgon', 'Gavroche', 'Gillenormand', 'Magnon',
                 'Mlle Gillenormand', 'Mme Pontmercy', 'Mlle Vaubois', 'Lt Gillenormand', 'Marius', 'Baroness',
                 'Mabeuf', 'Enjolras', 'Combeferre', 'Prouvaire', 'Feuilly', 'Courfeyrac', 'Bahorel', 'Bossuet',
                 'Joly', 'Grantaire', 'MotherPlutarch', 'Gueulemer', 'Babet', 'Claquesous', 'Montparnasse',
                 'Toussaint', 'Child1', 'Child2', 'Brujon', 'Mme Hucheloup']
        x = np.array(
            [0.53,  0.98,  0.41,  0.4,  1.,  0.92,  0.84,  0.74,  0.78, 1.,  0.51,  0.09, -0.,  0.29,  0.37,
             0.41, -0.35, -0.46, -0.42, -0.46, -0.41, -0.37, -0.36, -0.2, -0.06, -0.04, -0.01, -0.02,  0.33,
             0.17, -0.29, -0.1,  0.58,  0.29,  0.29,  0.26, 0.29,  0.37,  0.35,  0.04, -0.01, -0.18, -0.09,
             0.2,  0.51, 0.7, -0.95, -0.7, -0.37, -0.08, -0.18, -0.05,  0.04, -0.12, -0.06, -0.13, -0.24, -0.48,
             -0.25, -0.33, -0.43, -0.39, -0.33, -0.42, -0.31, -0.38, -0.48, -0.74, -0.08, -0.1, -0.02, -0.1,
             0.14, -0.76, -0.75, -0.18, -0.58])
        y = np.array(
            [-0.23, -0.42, -0.14, -0.18, -0.31, -0.52, -0.6, -0.65, -0.38, -0.19,  0.39,  0.03,  0.44, -0.44,
             0.51, -0.36,  0.27,  0.37, 0.4,  0.32,  0.32,  0.36,  0.4,  0.2,  0.07,  0.14, -0.05, 0.06,  0.06,
             0.24, -0.26, -0.1,  0.24, -0.04,  0.17,  0.23, 0.31,  0.21,  0.27, -0.36,  0.69,  0.11,  0.38, -0.09,
             0.05, 0.12,  0.82,  0.44,  0.06, -0.2, -0.4, -0.28, -0.68, -0.79, -0.4, -0.07, -0.51, -0.17, -0.03,
             -0.09, -0.14, -0.04, -0.04, -0.07, -0.06, -0.11, -0.06, -0.35,  0.24,  0.19,  0.22,  0.29, -0.2,
             0.06,  0.14,  0.3, -0.1])
        graph = Bunch()
        graph.adjacency = adjacency
        graph.names = np.array(names)
        graph.position = np.vstack((x, y)).T
        graph.name = 'miserables'
        return graph
    else:
        return adjacency


def painters(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Graph of links between some famous painters on Wikipedia.

    * Directed graph
    * 14 nodes, 50 edges
    * Names of painters

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    adjacency or graph : Union[sparse.csr_matrix, Bunch]
        Adjacency matrix or graph with metadata (names, positions).

    Example
    -------
    >>> from sknetwork.data import painters
    >>> adjacency = painters()
    >>> adjacency.shape
    (14, 14)
    """
    row = np.array(
        [0, 0, 1, 1, 2, 3, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5,
         6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 9, 9,
         10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13])
    col = np.array(
        [3, 10, 3, 12, 9, 0, 1, 7, 11, 12, 2, 5, 9, 2, 4, 8, 9,
         0, 13, 1, 2, 3, 8, 11, 12, 0, 1, 4, 5, 7, 10, 11, 2, 4,
         0, 3, 8, 11, 12, 0, 1, 3, 10, 12, 1, 3, 4, 7, 6, 8])
    adjacency = sparse.csr_matrix((np.ones(len(row), dtype=bool), (row, col)), shape=(14, 14))

    if metadata:
        names = np.array(
            ['Pablo Picasso', 'Claude Monet', 'Michel Angelo', 'Edouard Manet', 'Peter Paul Rubens', 'Rembrandt',
             'Gustav Klimt', 'Edgar Degas', 'Vincent van Gogh', 'Leonardo da Vinci', 'Henri Matisse', 'Paul Cezanne',
             'Pierre-Auguste Renoir', 'Egon Schiele'])
        x = np.array(
            [0.24, -0.47, -0.3, -0.31, -0.08, 0.12, 0.78, -0.36, 0.11,
             -0.06, -0.02, -0.12, -0.24, 0.73])
        y = np.array(
            [0.53, 0.19, -0.71, 0.44, -0.48, -0.65, 0.69, -0.11, 0.01,
             -1., 0.49, 0.28, 0.06, 0.27])
        graph = Bunch()
        graph.adjacency = adjacency
        graph.names = names
        graph.position = np.stack((x, y)).T
        graph.name = 'painters'
        return graph
    else:
        return adjacency


def hourglass(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Hourglass graph.

    * Bipartite graph
    * 4 nodes, 4 edges

    Returns
    -------
    biadjacency or graph : Union[sparse.csr_matrix, Bunch]
        Biadjacency matrix or graph.

    Example
    -------
    >>> from sknetwork.data import hourglass
    >>> biadjacency = hourglass()
    >>> biadjacency.shape
    (2, 2)
    """
    biadjacency = sparse.csr_matrix(np.ones((2, 2), dtype=bool))
    if metadata:
        graph = Bunch()
        graph.biadjacency = biadjacency
        return graph
    else:
        return biadjacency


def star_wars(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Bipartite graph connecting some Star Wars villains to the movies in which they appear.

    * Bipartite graph
    * 7 nodes (4 villains, 3 movies), 8 edges
    * Names of villains and movies

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    biadjacency or graph : Union[sparse.csr_matrix, Bunch]
        Biadjacency matrix or graph with metadata (names).

    Example
    -------
    >>> from sknetwork.data import star_wars
    >>> biadjacency = star_wars()
    >>> biadjacency.shape
    (4, 3)
   """
    row = np.array([0, 0, 1, 2, 2, 2, 3, 3])
    col = np.array([0, 2, 0, 0, 1, 2, 1, 2])
    biadjacency = sparse.csr_matrix((np.ones(len(row), dtype=bool), (row, col)), shape=(4, 3))

    if metadata:
        villains = np.array(['Jabba', 'Greedo', 'Vader', 'Boba'])
        movies = np.array(['A New Hope', 'The Empire Strikes Back', 'Return Of The Jedi'])
        graph = Bunch()
        graph.biadjacency = biadjacency
        graph.names = villains
        graph.names_row = villains
        graph.names_col = movies
        graph.name = 'star_wars'
        return graph
    else:
        return biadjacency


def movie_actor(metadata: bool = False) -> Union[sparse.csr_matrix, Bunch]:
    """Bipartite graph connecting movies to some actors starring in them.

    * Bipartite graph
    * 31 nodes (15 movies, 16 actors), 42 edges
    * 9 labels (rows)
    * Names of movies (rows) and actors (columns)
    * Names of movies production company (rows)

    Parameters
    ----------
    metadata :
        If ``True``, return a `Bunch` object with metadata.

    Returns
    -------
    biadjacency or graph : Union[sparse.csr_matrix, Bunch]
        Biadjacency matrix or graph with metadata (names).

    Example
    -------
    >>> from sknetwork.data import movie_actor
    >>> biadjacency = movie_actor()
    >>> biadjacency.shape
    (15, 16)
    """
    row = np.array(
        [0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 4, 4, 5, 5, 6,
         6, 6, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11,
         12, 12, 12, 13, 13, 14, 14])
    col = np.array(
        [0, 1, 2, 1, 2, 3, 3, 4, 5, 8, 4, 6, 0, 6, 4, 7, 4,
         7, 8, 3, 8, 9, 10, 11, 12, 15, 0, 11, 12, 9, 10, 13, 5, 9, 13,
         1, 9, 15, 12, 14, 11, 14])
    biadjacency = sparse.csr_matrix((np.ones(len(row), dtype=bool), (row, col)), shape=(15, 16))

    if metadata:
        movies = np.array(
            ['Inception', 'The Dark Knight Rises', 'The Big Short', 'Drive', 'The Great Gatsby', 'La La Land',
             'Crazy Stupid Love', 'Vice', 'The Grand Budapest Hotel', 'Aviator', '007 Spectre', 'Inglourious Basterds',
             'Midnight In Paris', 'Murder on the Orient Express', 'Fantastic Beasts 2'])
        actors = np.array(
            ['Leonardo DiCaprio', 'Marion Cotillard', 'Joseph Gordon Lewitt', 'Christian Bale', 'Ryan Gosling',
             'Brad Pitt', 'Carey Mulligan', 'Emma Stone', 'Steve Carell', 'Lea Seydoux', 'Ralph Fiennes', 'Jude Law',
             'Willem Dafoe', 'Christophe Waltz', 'Johnny Depp', 'Owen Wilson'])
        graph = Bunch()
        graph.biadjacency = biadjacency
        graph.names = movies
        graph.names_row = movies
        graph.names_col = actors
        graph.labels = np.array([0, 0, 1, 2, 3, 2, 4, 1, 5, 0, 6, 5, 7, 8, 0])
        graph.labels_name = np.array(['Warner Bros', 'Plan B Entertainment', 'Marc Platt Productions', 'Bazmark Films',
                                      'Carousel Productions', 'Babelsberg Studios', 'MGM', 'Gravier Productions',
                                      'Genre Films'])
        graph.labels_row = graph.labels
        graph.labels_row_name = graph.labels_name
        graph.name = 'movie_actor'
        return graph
    else:
        return biadjacency



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for data API"""

import tempfile
import unittest
import warnings

from sknetwork.data.load import *
from sknetwork.data.toy_graphs import *
from sknetwork.utils import Bunch


class TestDataAPI(unittest.TestCase):

    def test_toy_graphs(self):
        toy_graphs = [karate_club, painters, bow_tie, house, miserables]
        for toy_graph in toy_graphs:
            self.assertEqual(type(toy_graph()), sparse.csr_matrix)
            self.assertEqual(type(toy_graph(metadata=True)), Bunch)

    def test_load(self):
        tmp_data_dir = tempfile.gettempdir() + '/stub'
        clear_data_home(tmp_data_dir)
        try:
            graph = load_netset('stub', tmp_data_dir)
            self.assertEqual(type(graph), Bunch)
        except URLError:  # pragma: no cover
            warnings.warn('Could not reach NetSet. Corresponding test has not been performed.', RuntimeWarning)
            return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from abc import ABC

import numpy as np

from sknetwork.hierarchy.postprocess import split_dendrogram
from sknetwork.utils.base import Algorithm


class BaseHierarchy(Algorithm, ABC):
    """Base class for hierarchical clustering algorithms."""

    def __init__(self):
        self.dendrogram_ = None

    def fit_transform(self, *args, **kwargs) -> np.ndarray:
        """Fit algorithm to data and return the dendrogram. Same parameters as the ``fit`` method.

        Returns
        -------
        dendrogram : np.ndarray
            Dendrogram.
        """
        self.fit(*args, **kwargs)
        return self.dendrogram_


class BaseBiHierarchy(BaseHierarchy, ABC):
    """Base class for hierarchical clustering algorithms."""

    def __init__(self):
        super(BaseBiHierarchy, self).__init__()
        self.dendrogram_row_ = None
        self.dendrogram_col_ = None
        self.dendrogram_full_ = None

    def _split_vars(self, shape):
        dendrogram_row, dendrogram_col = split_dendrogram(self.dendrogram_, shape)

        self.dendrogram_full_ = self.dendrogram_
        self.dendrogram_ = dendrogram_row
        self.dendrogram_row_ = dendrogram_row
        self.dendrogram_col_ = dendrogram_col

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on November 19 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""
from typing import Union, Optional

import numpy as np
from scipy import sparse

from sknetwork.path.shortest_path import distance
from sknetwork.ranking.base import BaseRanking
from sknetwork.utils.check import check_format, check_square


class Harmonic(BaseRanking):
    """Harmonic centrality of each node in a connected graph, corresponding to the average inverse length of
    the shortest paths from that node to all the other ones.

    For a directed graph, the harmonic centrality is computed in terms of outgoing paths.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_jobs:
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.

    Attributes
    ----------
    scores_ : np.ndarray
        Harmonic centrality of each node.

    Example
    -------
    >>> from sknetwork.ranking import Harmonic
    >>> from sknetwork.data import house
    >>> harmonic = Harmonic()
    >>> adjacency = house()
    >>> scores = harmonic.fit_transform(adjacency)
    >>> np.round(scores, 2)
    array([3. , 3.5, 3. , 3. , 3.5])

    References
    ----------
    Marchiori, M., & Latora, V. (2000).
    `Harmony in the small-world.
    <https://arxiv.org/pdf/cond-mat/0008357.pdf>`_
    Physica A: Statistical Mechanics and its Applications, 285(3-4), 539-546.
    """

    def __init__(self, n_jobs: Optional[int] = None):
        super(Harmonic, self).__init__()

        self.n_jobs = n_jobs

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'Harmonic':
        """Harmonic centrality for connected graphs.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`Harmonic`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        n = adjacency.shape[0]
        indices = np.arange(n)

        dists = distance(adjacency, n_jobs=self.n_jobs, sources=indices)

        np.fill_diagonal(dists, 1)
        inv = (1 / dists)
        np.fill_diagonal(inv, 0)

        self.scores_ = inv.dot(np.ones(n))

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator


def diag_pinv(weights: np.ndarray) -> sparse.csr_matrix:
    """Compute :math:`W^+ = \\text{diag}(w)^+`, the pseudo inverse of the diagonal matrix
    with diagonal the weights :math:`w`.

    Parameters
    ----------
    weights:
        The weights to invert.

    Returns
    -------
    sparse.csr_matrix
        :math:`W^+`

    """
    diag: sparse.csr_matrix = sparse.diags(weights, format='csr')
    diag.data = 1 / diag.data
    return diag


def normalize(matrix: Union[sparse.csr_matrix, np.ndarray, LinearOperator], p=1):
    """Normalize rows of a matrix. Null rows remain null.

    Parameters
    ----------
    matrix :
        Input matrix.
    p :
        Order of the norm

    Returns
    -------
    normalized matrix : same as input
    """
    if p == 1:
        norm = matrix.dot(np.ones(matrix.shape[1]))
    elif p == 2:
        if isinstance(matrix, np.ndarray):
            norm = np.linalg.norm(matrix, axis=1)
        elif isinstance(matrix, sparse.csr_matrix):
            data = matrix.data.copy()
            matrix.data = data ** 2
            norm = np.sqrt(matrix.dot(np.ones(matrix.shape[1])))
            matrix.data = data
        else:
            raise NotImplementedError('Norm 2 is not available for LinearOperator.')
    else:
        raise NotImplementedError('Only norms 1 and 2 are available at the moment.')

    diag = diag_pinv(norm)
    if hasattr(matrix, 'left_sparse_dot') and callable(matrix.left_sparse_dot):
        return matrix.left_sparse_dot(diag)
    return diag.dot(matrix)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Louvain"""
import unittest

from sknetwork.clustering import Louvain, BiLouvain
from sknetwork.data import karate_club, star_wars
from sknetwork.data.test_graphs import *
from sknetwork.utils import bipartite2undirected


class TestLouvainClustering(unittest.TestCase):

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        n = adjacency.shape[0]
        labels = Louvain().fit_transform(adjacency)
        self.assertEqual(len(labels), n)

    def test_modularity(self):
        adjacency = karate_club()
        louvain_d = Louvain(modularity='dugue')
        louvain_n = Louvain(modularity='newman')
        labels_d = louvain_d.fit_transform(adjacency)
        labels_n = louvain_n.fit_transform(adjacency)
        self.assertTrue((labels_d == labels_n).all())

        louvain_p = Louvain(modularity='potts')
        louvain_p.fit_transform(adjacency)

    def test_bilouvain(self):
        biadjacency = star_wars()
        adjacency = bipartite2undirected(biadjacency)

        louvain = Louvain(modularity='newman')
        bilouvain = BiLouvain(modularity='newman')

        labels1 = louvain.fit_transform(adjacency)
        bilouvain.fit(biadjacency)
        labels2 = np.concatenate((bilouvain.labels_row_, bilouvain.labels_col_))
        self.assertTrue((labels1 == labels2).all())

    def test_options(self):
        adjacency = karate_club()

        # resolution
        louvain = Louvain(resolution=2)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 7)

        # tolerance
        louvain = Louvain(resolution=2, tol_aggregation=0.1)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 12)

        # shuffling
        louvain = Louvain(resolution=2, shuffle_nodes=True, random_state=42)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 9)

        # aggregate graph
        louvain = Louvain(return_aggregate=True)
        labels = louvain.fit_transform(adjacency)
        n_labels = len(set(labels))
        self.assertEqual(louvain.adjacency_.shape, (n_labels, n_labels))

        # aggregate graph
        Louvain(n_aggregations=1, sort_clusters=False).fit(adjacency)

    def test_options_with_64_bit(self):
        adjacency = karate_club()
        # force 64-bit index
        adjacency.indices = adjacency.indices.astype(np.int64)
        adjacency.indptr = adjacency.indptr.astype(np.int64)

        # resolution
        louvain = Louvain(resolution=2)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 7)

        # tolerance
        louvain = Louvain(resolution=2, tol_aggregation=0.1)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 12)

        # shuffling
        louvain = Louvain(resolution=2, shuffle_nodes=True, random_state=42)
        labels = louvain.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 9)

        # aggregate graph
        louvain = Louvain(return_aggregate=True)
        labels = louvain.fit_transform(adjacency)
        n_labels = len(set(labels))
        self.assertEqual(louvain.adjacency_.shape, (n_labels, n_labels))

        # aggregate graph
        Louvain(n_aggregations=1, sort_clusters=False).fit(adjacency)

        # check if labels are 64-bit
        self.assertEqual(labels.dtype, np.int64)

    def test_invalid(self):
        adjacency = karate_club()
        louvain = Louvain(modularity='toto')
        with self.assertRaises(ValueError):
            louvain.fit(adjacency)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 4 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.linalg.normalization import normalize


def projection_simplex_array(array: np.ndarray, scale: float = 1) -> np.ndarray:
    """Project each line of the input onto the Euclidean simplex i.e. solve

    :math:`\\underset{w}{min} ||w - x_i||_2^2` s.t. :math:`\\sum w_j = z, w_j \\ge 0`.

    Parameters
    ----------
    array: np.ndarray
        Data to project. Either one or two dimensional.
    scale: float
        Scale of the simplex i.e. sums of the projected coefficients.

    Returns
    -------
    projection : np.ndarray
        Array with the same shape as the input.

    Example
    -------
    >>> X = np.array([[2, 2], [-0.75, 0.25]])
    >>> projection_simplex_array(X)
    array([[0.5, 0.5],
           [0. , 1. ]])
    """
    if len(array.shape) == 1:
        array = array.reshape(1, array.shape[0])
    n_row, n_col = array.shape

    sorted_array = -np.sort(-array)
    cumsum_array = np.cumsum(sorted_array, axis=1) - scale
    denom = 1 + np.arange(n_col)
    condition = sorted_array - cumsum_array / denom > 0
    max_index = np.argmax(condition / denom[::-1], axis=1)
    threshold = (cumsum_array / denom)[np.arange(n_row), max_index]

    return np.maximum(array - threshold[:, np.newaxis], 0)


def projection_simplex_csr(matrix: sparse.csr_matrix, scale: float = 1):
    """Project each line of the input onto the Euclidean simplex i.e. solve

    :math:`\\underset{w}{min} ||w - x_i||_2^2` s.t. :math:`\\sum w_j = z, w_j \\ge 0`.

    Parameters
    ----------
    matrix : sparse.csr_matrix
        Matrix whose rows must be projected.
    scale: float
        Scale of the simplex i.e. sums of the projected coefficients.

    Returns
    -------
    projection : sparse.csr_matrix
        Matrix with the same shape as the input.

    Examples
    --------
    >>> X = sparse.csr_matrix(np.array([[2, 2], [-0.75, 0.25]]))
    >>> X_proj = projection_simplex_csr(X)
    >>> X_proj.nnz
    3
    >>> X_proj.toarray()
    array([[0.5, 0.5],
           [0. , 1. ]])
    """
    data = matrix.data
    if data.dtype == bool or (data.min() == data.max()):
        return normalize(matrix, p=1)

    indptr = matrix.indptr
    new_data = np.zeros_like(data)

    for i in range(indptr.size-1):
        j1 = indptr[i]
        j2 = indptr[i+1]
        new_data[j1:j2] = projection_simplex_array(data[j1:j2], scale=scale)

    new_matrix = sparse.csr_matrix((new_data, matrix.indices, indptr), shape=matrix.shape)
    new_matrix.eliminate_zeros()
    return new_matrix


def projection_simplex(x: Union[np.ndarray, sparse.csr_matrix], scale: float = 1.):
    """Project each line of the input onto the Euclidean simplex i.e. solve

    :math:`\\underset{w}{min} ||w - x_i||_2^2` s.t. :math:`\\sum w_j = z, w_j \\ge 0`.

    Parameters
    ----------
    x :
        Data to project. Either one or two dimensional. Can be sparse or dense.
    scale : float
        Scale of the simplex i.e. sums of the projected coefficients.

    Returns
    -------
    projection : np.ndarray or sparse.csr_matrix
        Array with the same type and shape as the input.

    Example
    -------
    >>> X = np.array([[2, 2], [-0.75, 0.25]])
    >>> projection_simplex(X)
    array([[0.5, 0.5],
           [0. , 1. ]])
    >>> X_csr = sparse.csr_matrix(X)
    >>> X_proj = projection_simplex(X_csr)
    >>> X_proj.nnz
    3
    >>> X_proj.toarray()
    array([[0.5, 0.5],
           [0. , 1. ]])

    References
    ----------
    Duchi, J., Shalev-Shwartz, S., Singer, Y., & Chandra, T. (2008, July).
    `Efficient projections onto the l 1-ball for learning in high dimensions.
    <http://machinelearning.org/archive/icml2008/papers/361.pdf>`_
    In Proceedings of the 25th international conference on Machine learning (pp. 272-279). ACM.
    """
    if isinstance(x, np.ndarray):
        return projection_simplex_array(x, scale)
    elif isinstance(x, sparse.csr_matrix):
        return projection_simplex_csr(x, scale)
    else:
        raise TypeError('Input must be a numpy array or a CSR matrix.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Weisfeiler-Lehman kernels"""
import unittest

import numpy as np

from sknetwork.data import house, bow_tie, linear_graph
from sknetwork.topology import are_isomorphic


class TestWLKernel(unittest.TestCase):

    def test_isomorphism(self):
        ref = house()
        n = ref.shape[0]

        adjacency = house()
        reorder = list(range(n))
        np.random.shuffle(reorder)
        adjacency = adjacency[reorder][:, reorder]
        self.assertTrue(are_isomorphic(ref, adjacency))

        adjacency = bow_tie()
        self.assertFalse(are_isomorphic(ref, adjacency))

        adjacency = linear_graph(n)
        self.assertFalse(are_isomorphic(ref, adjacency))

        adjacency = linear_graph(n + 1)
        self.assertFalse(are_isomorphic(ref, adjacency))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for visualization/graphs.py"""

import unittest

import numpy as np
from scipy import sparse

from sknetwork.data.test_graphs import test_graph_disconnect, test_bigraph_disconnect
from sknetwork.data.toy_graphs import karate_club, painters, movie_actor, bow_tie, star_wars
from sknetwork.visualization.graphs import svg_graph, svg_digraph, svg_bigraph


# noinspection DuplicatedCode
class TestVisualization(unittest.TestCase):

    def test_undirected(self):
        graph = karate_club(True)
        adjacency = graph.adjacency
        position = graph.position
        labels = graph.labels
        image = svg_graph(adjacency, position, labels=labels)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position, display_edges=False)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position, height=None)
        self.assertEqual(image[1:4], 'svg')
        n = adjacency.shape[0]
        edge_labels = [(0, 1, 0), (1, 1, 1), (3, 10, 2)]
        image = svg_graph(adjacency, position=None, names=np.arange(n), labels=np.arange(n), scores=np.arange(n),
                          seeds=[0, 1], width=200, height=200, margin=10, margin_text=5, scale=3,
                          node_order=np.flip(np.arange(n)),
                          node_size=5, node_size_min=2, node_size_max=6, display_node_weight=True,
                          node_weights=np.arange(n),
                          node_width=2, node_width_max=5, node_color='red', edge_width=2, edge_width_min=2,
                          edge_width_max=4, edge_color='blue', edge_labels=edge_labels, display_edge_weight=True,
                          font_size=14)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position=None, labels={0: 0})
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position=None, scores={0: 0})
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency=None, position=position)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency=None, position=position, edge_labels=edge_labels)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position, labels, label_colors={0: "red", 1: "blue"})
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position, labels, label_colors=["red", "blue"])
        self.assertEqual(image[1:4], 'svg')

    def test_directed(self):
        graph = painters(True)
        adjacency = graph.adjacency
        position = graph.position
        names = graph.names
        image = svg_digraph(adjacency, position, names=names)
        self.assertEqual(image[1:4], 'svg')
        image = svg_graph(adjacency, position, display_edges=False)
        self.assertEqual(image[1:4], 'svg')
        n = adjacency.shape[0]
        image = svg_digraph(adjacency, position=None, names=np.arange(n), labels=np.arange(n), scores=np.arange(n),
                            seeds=[0, 1], width=200, height=200, margin=10, margin_text=5, scale=3,
                            node_order=np.flip(np.arange(n)),
                            node_size=5, node_size_min=2, node_size_max=6, display_node_weight=True,
                            node_weights=np.arange(n),
                            node_width=2, node_width_max=5, node_color='red', edge_width=2, edge_width_min=2,
                            edge_width_max=4, edge_color='blue', display_edge_weight=True, font_size=14)
        self.assertEqual(image[1:4], 'svg')
        image = svg_digraph(adjacency, position=None, names=np.arange(n), labels={0: 1, 1: 2}, scores=np.arange(n),
                            seeds=[0, 1], width=200, height=200, margin=10, margin_text=5, scale=3,
                            node_order=np.flip(np.arange(n)), node_size=5, node_size_min=2, node_size_max=6,
                            display_node_weight=True, node_weights=np.arange(n), node_width=2, node_width_max=5,
                            node_color='red', edge_width=2, edge_width_min=2, edge_width_max=4, edge_color='blue',
                            display_edge_weight=True, font_size=14)
        self.assertEqual(image[1:4], 'svg')

    def test_bipartite(self):
        graph = movie_actor(True)
        biadjacency = graph.biadjacency
        names_row = graph.names_row
        names_col = graph.names_col
        image = svg_bigraph(biadjacency, names_row, names_col)
        self.assertEqual(image[1:4], 'svg')
        image = svg_bigraph(biadjacency, display_edges=False)
        self.assertEqual(image[1:4], 'svg')
        image = svg_bigraph(biadjacency, reorder=False)
        self.assertEqual(image[1:4], 'svg')
        n_row, n_col = biadjacency.shape
        position_row = np.random.random((n_row, 2))
        position_col = np.random.random((n_col, 2))
        edge_labels = [(0, 1, 0), (1, 1, 1), (3, 10, 2)]
        image = svg_bigraph(biadjacency=biadjacency, names_row=np.arange(n_row), names_col=np.arange(n_col),
                            labels_row=np.arange(n_row), labels_col=np.arange(n_col), scores_row=np.arange(n_row),
                            scores_col=np.arange(n_col), seeds_row=[0, 1], seeds_col=[1, 2],
                            position_row=position_row, position_col=position_col, color_row='red', color_col='white',
                            width=200, height=200, margin=10, margin_text=5, scale=3, node_size=5,
                            node_size_min=1, node_size_max=30, node_weights_row=np.arange(n_row),
                            node_weights_col=np.arange(n_col), display_node_weight=True, node_width=2, node_width_max=5,
                            edge_labels=edge_labels, edge_width=2, edge_width_min=0.3, edge_width_max=4,
                            edge_color='red', display_edge_weight=True, font_size=14)
        self.assertEqual(image[1:4], 'svg')

    def test_disconnect(self):
        adjacency = test_graph_disconnect()
        position = np.random.random((adjacency.shape[0], 2))
        image = svg_graph(adjacency, position)
        self.assertEqual(image[1:4], 'svg')
        biadjacency = test_bigraph_disconnect()
        image = svg_bigraph(biadjacency)
        self.assertEqual(image[1:4], 'svg')

    def test_membership(self):
        adjacency = bow_tie()
        membership = sparse.csr_matrix([[.5, .5], [0, 0], [1, 0], [0, 1], [0, 1]])
        image = svg_graph(adjacency, membership=membership)
        self.assertEqual(image[1:4], 'svg')
        biadjacency = star_wars()
        membership_row = sparse.csr_matrix([[.5, .5], [0, 0], [1, 0], [0, 1]])
        membership_col = sparse.csr_matrix([[.5, .5], [0, 0], [1, 0]])
        image = svg_bigraph(biadjacency, membership_row=membership_row, membership_col=membership_col)
        self.assertEqual(image[1:4], 'svg')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on July 17 2019
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union, Optional

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import bicgstab, LinearOperator

from sknetwork.linalg.normalization import normalize
from sknetwork.ranking.base import BaseRanking, BaseBiRanking
from sknetwork.utils.check import check_format, check_seeds, check_square, check_is_proba
from sknetwork.utils.format import bipartite2undirected
from sknetwork.utils.seeds import stack_seeds
from sknetwork.utils.verbose import VerboseMixin


class DirichletOperator(LinearOperator):
    """Diffusion in discrete time as a LinearOperator.

    Parameters
    ----------
    adjacency : sparse.csr_matrix
        Adjacency matrix of the graph.
    damping_factor : float
        Damping factor.
    border : np.ndarray (bool)
        Border nodes. If ``None``, then the diffusion is free.

    Attributes
    ----------
    a : sparse.csr_matrix
        Diffusion matrix.
    b : np.ndarray
        Regularization (uniform).
    """
    def __init__(self, adjacency: sparse.csr_matrix, damping_factor: float, border: np.ndarray = None):
        super(DirichletOperator, self).__init__(shape=adjacency.shape, dtype=float)
        n = adjacency.shape[0]
        out_nodes = adjacency.dot(np.ones(n)).astype(bool)
        if border is None:
            border = np.zeros(n, dtype=bool)
        interior: sparse.csr_matrix = sparse.diags(~border, shape=(n, n), format='csr', dtype=float)
        self.a = damping_factor * interior.dot(normalize(adjacency))
        self.b = interior.dot(np.ones(n) - damping_factor * out_nodes) / n

    def _matvec(self, x: np.ndarray):
        return self.a.dot(x) + self.b * x.sum()


class DeltaDirichletOperator(DirichletOperator):
    """Diffusion in discrete time as a LinearOperator (delta of temperature).

    Parameters
    ----------
    adjacency : sparse.csr_matrix
        Adjacency matrix of the graph.
    damping_factor : float
        Damping factor.
    border : np.ndarray (bool)
        Border nodes. If ``None``, then the diffusion is free.

    Attributes
    ----------
    a : sparse.csr_matrix
        Diffusion matrix.
    b : np.ndarray
        Regularization (uniform).
    """
    def __init__(self, adjacency: sparse.csr_matrix, damping_factor: float, border: np.ndarray = None):
        super(DeltaDirichletOperator, self).__init__(adjacency, damping_factor, border)

    def _matvec(self, x: np.ndarray):
        return self.a.dot(x) + self.b * x.sum() - x


class Diffusion(BaseRanking):
    """Ranking by diffusion along the edges (heat equation).

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_iter : int
        Number of steps of the diffusion in discrete time (must be positive).
    damping_factor : float (optional)
        Damping factor (default value = 1).

    Attributes
    ----------
    scores_ : np.ndarray
        Score of each node (= temperature).

    Example
    -------
    >>> from sknetwork.data import house
    >>> diffusion = Diffusion(n_iter=2)
    >>> adjacency = house()
    >>> seeds = {0: 1, 2: 0}
    >>> scores = diffusion.fit_transform(adjacency, seeds)
    >>> np.round(scores, 2)
    array([0.58, 0.56, 0.38, 0.58, 0.42])

    References
    ----------
    Chung, F. (2007). The heat kernel as the pagerank of a graph. Proceedings of the National Academy of Sciences.
    """
    def __init__(self, n_iter: int = 3, damping_factor: Optional[float] = None):
        super(Diffusion, self).__init__()

        if n_iter <= 0:
            raise ValueError('The number of iterations must be positive.')
        else:
            self.n_iter = n_iter
        if damping_factor is None:
            damping_factor = 1.
        check_is_proba(damping_factor, 'Damping factor')
        self.damping_factor = damping_factor

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds: Optional[Union[dict, np.ndarray]] = None, init: Optional[float] = None) \
            -> 'Diffusion':
        """Compute the diffusion (temperatures at equilibrium).

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.
        seeds :
            Temperatures of seed nodes in initial state (dictionary or vector). Negative temperatures ignored.
        init :
            Temperature of non-seed nodes in initial state.
            If ``None``, use the average temperature of seed nodes (default).

        Returns
        -------
        self: :class:`Diffusion`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        n: int = adjacency.shape[0]
        if seeds is None:
            self.scores_ = np.ones(n) / n
            return self

        seeds = check_seeds(seeds, n)
        border = (seeds >= 0)

        if init is None:
            scores = seeds[border].mean() * np.ones(n)
        else:
            scores = init * np.ones(n)
        scores[border] = seeds[border]

        diffusion = DirichletOperator(adjacency, self.damping_factor)
        for i in range(self.n_iter):
            scores = diffusion.dot(scores)

        self.scores_ = scores

        return self


class BiDiffusion(Diffusion, BaseBiRanking):
    """Ranking by diffusion along the edges of a bipartite graph (heat equation).

    * Bigraphs

    Parameters
    ----------
    n_iter : int
        Number of steps of the diffusion in discrete time (must be positive).
    damping_factor : float (optional)
        Damping factor (default value = 1).

    Attributes
    ----------
    scores_ : np.ndarray
        Scores of rows.
    scores_row_ : np.ndarray
        Scores of rows (copy of **scores_**).
    scores_col_ : np.ndarray
        Scores of columns.

    Example
    -------
    >>> from sknetwork.ranking import BiDiffusion
    >>> from sknetwork.data import star_wars
    >>> bidiffusion = BiDiffusion(n_iter=2)
    >>> biadjacency = star_wars()
    >>> scores = bidiffusion.fit_transform(biadjacency, seeds_row = {0: 1, 2: 0})
    >>> np.round(scores, 2)
    array([0.5 , 0.5 , 0.46, 0.44])
    """
    def __init__(self, n_iter: int = 3, damping_factor: Optional[float] = None):
        super(BiDiffusion, self).__init__(n_iter, damping_factor)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds_row: Optional[Union[dict, np.ndarray]] = None, seeds_col: Optional[Union[dict, np.ndarray]] = None,
            init: Optional[float] = None) -> 'BiDiffusion':
        """Compute the diffusion (temperatures at equilibrium).

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix, shape (n_row, n_col).
        seeds_row :
            Temperatures of seed rows in initial state (dictionary or vector of size n_row).
            Negative temperatures ignored.
        seeds_col :
            Temperatures of seed columns  in initial state (dictionary or vector of size n_col).
            Negative temperatures ignored.
        init :
            Temperature of non-seed nodes in initial state.
            If ``None``, use the average temperature of seed nodes (default).
        Returns
        -------
        self: :class:`BiDiffusion`
        """
        biadjacency = check_format(biadjacency)
        n_row, n_col = biadjacency.shape
        seeds = stack_seeds(n_row, n_col, seeds_row, seeds_col)
        adjacency = bipartite2undirected(biadjacency)
        Diffusion.fit(self, adjacency, seeds, init)
        # average over 2 successive iterations because the graph is bipartite
        diffusion = DirichletOperator(adjacency, self.damping_factor)
        self.scores_ += diffusion.dot(self.scores_)
        self.scores_ /= 2
        self._split_vars(n_row)

        return self


class Dirichlet(BaseRanking, VerboseMixin):
    """Ranking by the Dirichlet problem (heat diffusion with boundary constraints).

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_iter : int
        If positive, number of steps of the diffusion in discrete time.
        Otherwise, solve the Dirichlet problem by the bi-conjugate gradient stabilized method.
    damping_factor : float (optional)
        Damping factor (default value = 1).
    verbose : bool
        Verbose mode.

    Attributes
    ----------
    scores_ : np.ndarray
        Score of each node (= temperature).

    Example
    -------
    >>> from sknetwork.ranking import Dirichlet
    >>> from sknetwork.data import house
    >>> dirichlet = Dirichlet()
    >>> adjacency = house()
    >>> seeds = {0: 1, 2: 0}
    >>> scores = dirichlet.fit_transform(adjacency, seeds)
    >>> np.round(scores, 2)
    array([1.  , 0.54, 0.  , 0.31, 0.62])

    References
    ----------
    Chung, F. (2007). The heat kernel as the pagerank of a graph. Proceedings of the National Academy of Sciences.
    """
    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, verbose: bool = False):
        super(Dirichlet, self).__init__()
        VerboseMixin.__init__(self, verbose)

        self.n_iter = n_iter
        if damping_factor is None:
            damping_factor = 1.
        check_is_proba(damping_factor, 'Damping factor')
        self.damping_factor = damping_factor

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds: Optional[Union[dict, np.ndarray]] = None, init: Optional[float] = None) -> 'Dirichlet':
        """Compute the solution to the Dirichlet problem (temperatures at equilibrium).

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.
        seeds :
            Temperatures of seed nodes (dictionary or vector). Negative temperatures ignored.
        init :
            Temperature of non-seed nodes in initial state.
            If ``None``, use the average temperature of seed nodes (default).

        Returns
        -------
        self: :class:`Dirichlet`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        n: int = adjacency.shape[0]
        if seeds is None:
            self.scores_ = np.ones(n) / n
            return self

        seeds = check_seeds(seeds, n)
        border = (seeds >= 0)

        if init is None:
            scores = seeds[border].mean() * np.ones(n)
        else:
            scores = init * np.ones(n)
        scores[border] = seeds[border]

        if self.n_iter > 0:
            diffusion = DirichletOperator(adjacency, self.damping_factor, border)
            for i in range(self.n_iter):
                scores = diffusion.dot(scores)
                scores[border] = seeds[border]
        else:
            a = DeltaDirichletOperator(adjacency, self.damping_factor, border)
            b = -seeds
            b[~border] = 0
            scores, info = bicgstab(a, b, atol=0., x0=scores)
            self._scipy_solver_info(info)

        tmin, tmax = seeds[border].min(), seeds[border].max()
        self.scores_ = np.clip(scores, tmin, tmax)

        return self


class BiDirichlet(Dirichlet, BaseBiRanking):
    """Ranking by the Dirichlet problem in bipartite graphs (heat diffusion with boundary constraints).

    * Bigraphs

    Parameters
    ----------
    n_iter : int
        If positive, number of steps of the diffusion in discrete time.
        Otherwise, solve the Dirichlet problem by the bi-conjugate gradient stabilized method.
    damping_factor : float (optional)
        Damping factor (default value = 1).
    verbose : bool
        Verbose mode.

    Attributes
    ----------
    scores_ : np.ndarray
        Scores of rows.
    scores_row_ : np.ndarray
        Scores of rows (copy of **scores_**).
    scores_col_ : np.ndarray
        Scores of columns.

    Example
    -------
    >>> from sknetwork.ranking import BiDirichlet
    >>> from sknetwork.data import star_wars
    >>> bidirichlet = BiDirichlet()
    >>> biadjacency = star_wars()
    >>> scores = bidirichlet.fit_transform(biadjacency, seeds_row = {0: 1, 2: 0})
    >>> np.round(scores, 2)
    array([1.  , 0.5 , 0.  , 0.29])
    """

    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, verbose: bool = False):
        super(BiDirichlet, self).__init__(n_iter, damping_factor, verbose)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds_row: Optional[Union[dict, np.ndarray]] = None, seeds_col: Optional[Union[dict, np.ndarray]] = None,
            init: Optional[float] = None) -> 'BiDirichlet':
        """Compute the solution to the Dirichlet problem (temperatures at equilibrium).

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix, shape (n_row, n_col).
        seeds_row :
            Temperatures of seed rows (dictionary or vector of size n_row). Negative temperatures ignored.
        seeds_col :
            Temperatures of seed columns (dictionary or vector of size n_col). Negative temperatures ignored.
        init :
            Temperature of non-seed nodes in initial state.
            If ``None``, use the average temperature of seed nodes (default).

        Returns
        -------
        self: :class:`BiDirichlet`
        """
        biadjacency = check_format(biadjacency)
        n_row, n_col = biadjacency.shape
        seeds = stack_seeds(n_row, n_col, seeds_row, seeds_col)
        adjacency = bipartite2undirected(biadjacency)
        Dirichlet.fit(self, adjacency, seeds, init)
        self._split_vars(n_row)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for seeds.py"""

import unittest

import numpy as np

from sknetwork.utils.seeds import stack_seeds, seeds2probs


class TestSeeds(unittest.TestCase):

    def test_seeds2probs(self):
        n = 4
        seeds_array = np.array([0, 1, -1, 0])
        seeds_dict = {0: 0, 1: 1, 3: 0}

        probs1 = seeds2probs(n, seeds_array)
        probs2 = seeds2probs(n, seeds_dict)
        self.assertTrue(np.allclose(probs1, probs2))

        bad_input = np.array([0, 0, -1, 0])
        with self.assertRaises(ValueError):
            seeds2probs(n, bad_input)

    def test_stack_seeds(self):
        n_row, n_col = 4, 3
        seeds_row_array = np.array([0, 1, -1, 0])
        seeds_row_dict = {0: 0, 1: 1, 3: 0}
        seeds_col_array = np.array([0, 1, -1])
        seeds_col_dict = {0: 0, 1: 1}

        seeds1 = stack_seeds(n_row, n_col, seeds_row_array, seeds_col_array)
        seeds2 = stack_seeds(n_row, n_col, seeds_row_dict, seeds_col_dict)
        seeds3 = stack_seeds(n_row, n_col, seeds_row_array, seeds_col_dict)
        seeds4 = stack_seeds(n_row, n_col, seeds_row_dict, seeds_col_array)

        self.assertTrue(np.allclose(seeds1, seeds2))
        self.assertTrue(np.allclose(seeds2, seeds3))
        self.assertTrue(np.allclose(seeds3, seeds4))

        seeds1 = stack_seeds(n_row, n_col, seeds_row_array, None)
        seeds2 = stack_seeds(n_row, n_col, seeds_row_dict, None)

        self.assertTrue(np.allclose(seeds1, seeds2))

        seeds1 = stack_seeds(n_col, n_row, None, seeds_row_array)
        seeds2 = stack_seeds(n_col, n_row, None, seeds_row_dict)

        self.assertTrue(np.allclose(seeds1, seeds2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Dec 3, 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""


class Log:
    """Log class for easier verbosity features"""
    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self.log = ''

    def print(self, *args):
        """Fill log with text."""
        if self.verbose:
            print(*args)
        self.log += ' '.join(map(str, args)) + '\n'

    def __repr__(self):
        return self.log


class VerboseMixin:
    """Mixin class for verbosity"""
    def __init__(self, verbose: bool = False):
        self.log = Log(verbose)

    def _scipy_solver_info(self, info: int):
        """Fill log with scipy info."""
        if info == 0:
            self.log.print('Successful exit.')
        elif info > 0:
            self.log.print('Convergence to tolerance not achieved.')
        else:
            self.log.print('Illegal input or breakdown.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2019
@author: Thomas Bonald <bonald@enst.fr>
@author: Quentin Lutz <qlutz@enst.fr>
"""

import unittest

from sknetwork.data import karate_club
from sknetwork.hierarchy import Paris, cut_straight, cut_balanced, aggregate_dendrogram


# noinspection PyMissingOrEmptyDocstring
class TestCuts(unittest.TestCase):

    def setUp(self):
        paris = Paris()
        adjacency = karate_club()
        self.dendrogram = paris.fit_transform(adjacency)

    def test_cuts(self):
        labels = cut_straight(self.dendrogram)
        self.assertEqual(len(set(labels)), 2)
        labels = cut_straight(self.dendrogram, n_clusters=5)
        self.assertEqual(len(set(labels)), 5)
        labels = cut_balanced(self.dendrogram, 2)
        self.assertEqual(len(set(labels)), 21)
        labels, new_dendrogram = cut_balanced(self.dendrogram, max_cluster_size=2, return_dendrogram=True)
        self.assertEqual(len(set(labels)), 21)
        self.assertTupleEqual(new_dendrogram.shape, (20, 4))

    def test_options(self):
        labels = cut_straight(self.dendrogram, threshold=0.5)
        self.assertEqual(len(set(labels)), 4)
        labels = cut_straight(self.dendrogram, n_clusters=3, threshold=0.5)
        self.assertEqual(len(set(labels)), 3)
        labels = cut_straight(self.dendrogram, sort_clusters=False)
        self.assertEqual(len(set(labels)), 2)
        labels = cut_balanced(self.dendrogram, max_cluster_size=2, sort_clusters=False)
        self.assertEqual(len(set(labels)), 21)
        labels = cut_balanced(self.dendrogram, max_cluster_size=10)
        self.assertEqual(len(set(labels)), 5)

    def test_aggregation(self):
        aggregated = aggregate_dendrogram(self.dendrogram, n_clusters=3)
        self.assertEqual(len(aggregated), 2)

        aggregated, counts = aggregate_dendrogram(self.dendrogram, n_clusters=3, return_counts=True)
        self.assertEqual(len(aggregated), 2)
        self.assertEqual(len(counts), 3)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 31 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np


def top_k(scores: np.ndarray, k: int = 1):
    """Index of the k elements with highest value.

    Parameters
    ----------
    scores : np.ndarray
        Array of values.
    k : int
        Number of elements to return.

    Examples
    --------
    >>> scores = np.array([0, 1, 0, 0.5])
    >>> top_k(scores, k=2)
    array([1, 3])

    Notes
    -----
    This is a basic implementation that sorts the entire array to find its top k elements.
    """
    return np.argsort(-scores)[:k]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for polynomes"""

import unittest

import numpy as np
from scipy import sparse

from sknetwork.data.test_graphs import test_graph
from sknetwork.linalg import Polynome, randomized_eig, randomized_svd


class TestPolynome(unittest.TestCase):

    def test_init(self):
        adjacency = test_graph()
        with self.assertRaises(ValueError):
            Polynome(adjacency, np.array([]))

    def test_operations(self):
        adjacency = test_graph()
        n = adjacency.shape[0]
        polynome = Polynome(adjacency, np.arange(3))
        x = np.random.randn(n)

        y1 = (polynome * 2).dot(x)
        y2 = (-polynome).dot(x)
        self.assertAlmostEqual(np.linalg.norm(0.5 * y1 + y2), 0)

    def test_dot(self):
        adjacency = sparse.eye(5, format='csr')
        polynome = Polynome(adjacency, np.arange(2))

        x = np.random.randn(5, 3)
        y = polynome.dot(x)
        self.assertAlmostEqual(np.linalg.norm(x - y), 0)

    def test_decomposition(self):
        adjacency = test_graph()
        n = adjacency.shape[0]
        polynome = Polynome(adjacency, np.arange(3))

        eigenvalues, eigenvectors = randomized_eig(polynome, n_components=2, which='LM')
        self.assertEqual(eigenvalues.shape, (2,))
        self.assertEqual(eigenvectors.shape, (n, 2))

        left_sv, sv, right_sv = randomized_svd(polynome, n_components=2)
        self.assertEqual(left_sv.shape, (n, 2))
        self.assertEqual(sv.shape, (2,))
        self.assertEqual(right_sv.shape, (2, n))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for parse.py"""

import unittest
from os import remove

import numpy as np

from sknetwork.data import parse


class TestParser(unittest.TestCase):

    def test_unlabeled_unweighted(self):
        self.stub_data_1 = 'stub_1.txt'
        with open(self.stub_data_1, "w") as text_file:
            text_file.write('%stub\n1 3\n4 5\n0 2')
        graph = parse.load_edge_list(self.stub_data_1)
        adjacency = graph.adjacency
        self.assertTrue((adjacency.indices == [2, 3, 0, 1, 5, 4]).all())
        self.assertTrue((adjacency.indptr == [0, 1, 2, 3, 4, 5, 6]).all())
        self.assertTrue((adjacency.data == [1, 1, 1, 1, 1, 1]).all())
        remove(self.stub_data_1)

    def test_labeled_weighted(self):
        self.stub_data_2 = 'stub_2.txt'
        with open(self.stub_data_2, "w") as text_file:
            text_file.write('%stub\nf, e, 5\na, d, 6\nc, b, 1')
        graph = parse.load_edge_list(self.stub_data_2)
        adjacency = graph.adjacency
        names = graph.names
        self.assertTrue((adjacency.indices == [4, 3, 5, 1, 0, 2]).all())
        self.assertTrue((adjacency.indptr == [0, 1, 2, 3, 4, 5, 6]).all())
        self.assertTrue((adjacency.data == [1, 6, 5, 6, 1, 5]).all())
        self.assertTrue((names == [' b', ' d', ' e', 'a', 'c', 'f']).all())
        remove(self.stub_data_2)

    def test_auto_reindex(self):
        self.stub_data_4 = 'stub_4.txt'
        with open(self.stub_data_4, "w") as text_file:
            text_file.write('%stub\n14 31\n42 50\n0 12')
        graph = parse.load_edge_list(self.stub_data_4)
        adjacency = graph.adjacency
        names = graph.names
        self.assertTrue((adjacency.indices == [1, 0, 3, 2, 5, 4]).all())
        self.assertTrue((adjacency.indptr == [0, 1, 2, 3, 4, 5, 6]).all())
        self.assertTrue((adjacency.data == [1, 1, 1, 1, 1, 1]).all())
        self.assertTrue((names == [0, 12, 14, 31, 42, 50]).all())
        remove(self.stub_data_4)

    def test_wrong_format_fast(self):
        self.stub_data_3 = 'stub_3.txt'
        with open(self.stub_data_3, "w") as text_file:
            text_file.write('%stub\n1 3 a\n4 5 b\n0 2 e')
        self.assertRaises(ValueError, parse.load_edge_list, self.stub_data_3)
        remove(self.stub_data_3)

    def test_wrong_format_slow(self):
        self.stub_data_3 = 'stub_3.txt'
        with open(self.stub_data_3, "w") as text_file:
            text_file.write('%stub\n1 3 a\n4 5 b\n0 2 e')
        self.assertRaises(ValueError, parse.load_edge_list, self.stub_data_3, fast_format=False)
        remove(self.stub_data_3)

    def test_graphml_basic(self):
        self.stub_data_5 = 'stub_5.graphml'
        with open(self.stub_data_5, "w") as graphml_file:
            graphml_file.write("""<?xml version='1.0' encoding='utf-8'?>
                                    <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                                    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
                                    http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
                                    <key id="d0" for="edge" attr.name="weight" attr.type="int"/>
                                    <graph edgedefault="directed">
                                    <node id="node1"/>
                                    <node id="node2"/>
                                    <edge source="node1" target="node2">
                                      <data key="d0">1</data>
                                    </edge></graph></graphml>""")
        graph = parse.load_graphml(self.stub_data_5)
        adjacency = graph.adjacency
        names = graph.names
        self.assertTrue((adjacency.indices == [1]).all())
        self.assertTrue((adjacency.indptr == [0, 1, 1]).all())
        self.assertTrue((adjacency.data == [1]).all())
        self.assertTrue((names == ['node1', 'node2']).all())
        remove(self.stub_data_5)

    def test_graphml_refined(self):
        self.stub_data_6 = 'stub_6.graphml'
        with open(self.stub_data_6, "w") as graphml_file:
            graphml_file.write("""<?xml version='1.0' encoding='utf-8'?>
                                    <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                                    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
                                    http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
                                      <desc>Some file</desc>
                                    <key id="d0" for="edge" attr.name="weight" attr.type="int"/>
                                    <key id="d1" for="node" attr.name="color" attr.type="string">
                                      <desc>Color</desc>
                                      <default>blue</default>
                                    </key>
                                    <key id="d2" for="edge" attr.name="distance" attr.type="double">
                                      <desc>Distance</desc>
                                      <default>1.5</default>
                                    </key>
                                    <graph edgedefault="undirected" parse.nodeids="canonical"
                                    parse.nodes="3" parse.edges="4">
                                    <node id="n0">
                                      <data key="d1">green</data>
                                    </node>
                                    <node id="n1"/>
                                    <node id="n2"/>
                                    <edge source="n0" target="n1">
                                      <data key="d0">1</data>
                                      <data key="d2">7.2</data>
                                    </edge>
                                    <edge source="n1" target="n2" directed='true'>
                                      <data key="d0">1</data>
                                    </edge>
                                    <edge source="n0" target="n2" directed='false'>
                                      <data key="d0">1</data>
                                    </edge></graph></graphml>""")
        graph = parse.load_graphml(self.stub_data_6)
        adjacency = graph.adjacency
        colors = graph.node_attribute.color
        distances = graph.edge_attribute.distance
        self.assertTrue((adjacency.indices == [1, 2, 0, 2, 0]).all())
        self.assertTrue((adjacency.indptr == [0, 2, 4, 5]).all())
        self.assertTrue((adjacency.data == [1, 1, 1, 1, 1]).all())
        self.assertTrue((colors == ['green', 'blue', 'blue']).all())
        self.assertTrue((distances == [7.2, 7.2, 1.5, 1.5, 1.5]).all())
        self.assertEqual(graph.meta.description, 'Some file')
        self.assertEqual(graph.meta.attributes.node.color, 'Color')
        self.assertEqual(graph.meta.attributes.edge.distance, 'Distance')
        remove(self.stub_data_6)

    def test_no_graphml(self):
        self.stub_data_7 = 'stub_7.graphml'
        with open(self.stub_data_7, "w") as graphml_file:
            graphml_file.write("""<?xml version='1.0' encoding='utf-8'?>
                                    <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                                    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
                                    http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
                                    <key id="d0" for="edge" attr.name="weight" attr.type="int"/>
                                    </graphml>""")
        self.assertRaises(ValueError, parse.load_graphml, self.stub_data_7)
        remove(self.stub_data_7)

    def test_adjacency_list(self):
        self.stub_data_8 = 'stub_8.txt'
        with open(self.stub_data_8, "w") as text_file:
            text_file.write('%stub\n2\n3\n0\n1\n5\n4')
        graph = parse.load_adjacency_list(self.stub_data_8)
        adjacency = graph.adjacency
        self.assertTupleEqual(adjacency.shape, (6, 6))
        self.assertTrue((adjacency.indices == [2, 3, 0, 1, 5, 4]).all())
        self.assertTrue((adjacency.indptr == [0, 1, 2, 3, 4, 5, 6]).all())
        self.assertTrue((adjacency.data == [1, 1, 1, 1, 1, 1]).all())
        remove(self.stub_data_8)

    def test_bipartite_edge_list(self):
        self.stub_data_9 = 'stub_9.txt'
        with open(self.stub_data_9, "w") as text_file:
            text_file.write('%stub\n1 3\n4 5\n0 3')
        graph = parse.load_edge_list(self.stub_data_9, bipartite=True)
        biadjacency = graph.biadjacency
        self.assertTrue((biadjacency.indices == [0, 0, 1]).all())
        self.assertTrue((biadjacency.indptr == [0, 1, 2, 3]).all())
        self.assertTrue((biadjacency.data == [1, 1, 1]).all())
        remove(self.stub_data_9)

    def test_bipartite_adjacency_list(self):
        self.stub_data_10 = 'stub_10.txt'
        with open(self.stub_data_10, "w") as text_file:
            text_file.write('%stub\n3\n3\n0')
        graph = parse.load_adjacency_list(self.stub_data_10, bipartite=True)
        biadjacency = graph.biadjacency
        self.assertTupleEqual(biadjacency.shape, (3, 4))
        self.assertTrue((biadjacency.indices == [3, 3, 0]).all())
        self.assertTrue((biadjacency.indptr == [0, 1, 2, 3]).all())
        self.assertTrue((biadjacency.data == [1, 1, 1]).all())
        remove(self.stub_data_10)

    def test_local_edge_list(self):
        edge_list_1 = [('Alice', 'Bob'), ('Carol', 'Alice')]
        graph = parse.convert_edge_list(edge_list_1)
        adjacency = graph.adjacency
        names = graph.names
        self.assertTrue((names == ['Alice', 'Bob', 'Carol']).all())
        self.assertTupleEqual(adjacency.shape, (3, 3))
        self.assertTrue((adjacency.indptr == [0, 2, 3, 4]).all())
        self.assertTrue((adjacency.indices == [1, 2, 0, 0]).all())
        self.assertTrue((adjacency.data == [1, 1, 1, 1]).all())

        edge_list_2 = [('Alice', 'Bob', 4), ('Carol', 'Alice', 6)]
        graph = parse.convert_edge_list(edge_list_2)
        adjacency = graph.adjacency
        names = graph.names
        self.assertTrue((names == ['Alice', 'Bob', 'Carol']).all())
        self.assertTupleEqual(adjacency.shape, (3, 3))
        self.assertTrue((adjacency.indptr == [0, 2, 3, 4]).all())
        self.assertTrue((adjacency.indices == [1, 2, 0, 0]).all())
        self.assertTrue((adjacency.data == [4, 6, 4, 6]).all())

    def test_bad_format_edge_list(self):
        edge_list_1 = [('Alice', 'Bob', 3, 5), ('Carol', 'Alice', 6, 8)]
        self.assertRaises(ValueError, parse.convert_edge_list, edge_list_1)
        edge_list_2 = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        self.assertRaises(ValueError, parse.convert_edge_list, edge_list_2)
        edge_list_3 = 'ab cd'
        self.assertRaises(TypeError, parse.convert_edge_list, edge_list_3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for closeness.py"""

import unittest

from sknetwork.data.test_graphs import *
from sknetwork.ranking.closeness import Closeness


class TestDiffusion(unittest.TestCase):

    def test_params(self):
        with self.assertRaises(ValueError):
            adjacency = test_graph()
            Closeness(method='toto').fit(adjacency)

    def test_parallel(self):
        adjacency = test_graph()
        n = adjacency.shape[0]

        closeness = Closeness(method='approximate')
        scores1 = closeness.fit_transform(adjacency)
        closeness = Closeness(method='approximate', n_jobs=-1)
        scores2 = closeness.fit_transform(adjacency)

        self.assertEqual(scores1.shape, (n,))
        self.assertAlmostEqual(np.linalg.norm(scores1 - scores2), 0)

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        closeness = Closeness()
        with self.assertRaises(ValueError):
            closeness.fit(adjacency)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for diffusion.py"""

import unittest

from sknetwork.data.test_graphs import test_bigraph
from sknetwork.ranking import HITS


class TestHITS(unittest.TestCase):

    def test_keywords(self):
        biadjacency = test_bigraph()
        n_row, n_col = biadjacency.shape

        for hits in [HITS(solver='lanczos', tol=0.1), HITS(solver='halko', n_oversamples=1)]:
            hits.fit(biadjacency)
            self.assertEqual(hits.scores_row_.shape, (n_row,))
            self.assertEqual(hits.scores_col_.shape, (n_col,))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2020
@author: Quentin Lutz <qlutz@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""

import unittest

from sknetwork.data.test_graphs import test_graph
from sknetwork.hierarchy.louvain_hierarchy import LouvainHierarchy


class TestLouvainHierarchy(unittest.TestCase):

    def test_options(self):
        louvain = LouvainHierarchy(resolution=2)
        adjacency = test_graph()
        dendrogram = louvain.fit_transform(adjacency)
        n = adjacency.shape[0]
        self.assertEqual(dendrogram.shape, (n - 1, 4))

        louvain = LouvainHierarchy(depth=1)
        adjacency = test_graph()
        dendrogram = louvain.fit_transform(adjacency)
        n = adjacency.shape[0]
        self.assertEqual(dendrogram.shape, (n - 1, 4))
        self.assertEqual(dendrogram[:, 2].max(), 1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for check.py"""
import unittest

from sknetwork.data import cyclic_digraph
from sknetwork.data.test_graphs import test_graph_disconnect
from sknetwork.utils.check import *
from sknetwork.utils.format import check_csr_or_slr


class TestChecks(unittest.TestCase):

    def setUp(self):
        """Simple graphs for tests."""
        self.adjacency = cyclic_digraph(3)
        self.dense_mat = np.identity(3)

    def test_check_format(self):
        with self.assertRaises(TypeError):
            check_format(self.adjacency.tocsc())

    def test_check_csr_slr(self):
        with self.assertRaises(TypeError):
            check_csr_or_slr(np.ones(3))

    def test_check_square(self):
        with self.assertRaises(ValueError):
            check_square(np.ones((3, 7)))

    def test_check_connected(self):
        with self.assertRaises(ValueError):
            check_connected(test_graph_disconnect())

    def test_check_symmetry(self):
        with self.assertRaises(ValueError):
            check_symmetry(self.adjacency)

    def test_nonnegative_entries(self):
        self.assertTrue(has_nonnegative_entries(self.adjacency))
        self.assertTrue(has_nonnegative_entries(self.dense_mat))

    def test_check_nonnegative(self):
        with self.assertRaises(ValueError):
            check_nonnegative(-self.dense_mat)

    def test_positive_entries(self):
        self.assertFalse(has_positive_entries(self.dense_mat))
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            has_positive_entries(self.adjacency)

    def test_check_positive(self):
        check_positive(np.ones(3))
        with self.assertRaises(ValueError):
            check_positive(-self.dense_mat)

    def test_probas(self):
        self.assertTrue(is_proba_array(np.array([.5, .5])))
        check_is_proba(0.5)
        with self.assertRaises(TypeError):
            is_proba_array(np.ones((2, 2, 2)))
        self.assertRaises(TypeError, check_is_proba, 'toto')
        with self.assertRaises(ValueError):
            check_is_proba(2)

    def test_damping(self):
        with self.assertRaises(ValueError):
            check_damping_factor(1)

    def test_error_make_weights(self):
        with self.assertRaises(ValueError):
            make_weights(distribution='junk', adjacency=self.adjacency)

    def test_error_check_is_proba(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            check_is_proba('junk')
        with self.assertRaises(ValueError):
            check_is_proba(2)

    def test_error_check_weights(self):
        with self.assertRaises(ValueError):
            check_weights(np.zeros(4), self.adjacency)
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            check_weights(2, self.adjacency)
        with self.assertRaises(ValueError):
            check_weights(np.zeros(3), self.adjacency, positive_entries=True)
        with self.assertRaises(ValueError):
            check_weights(-np.ones(3), self.adjacency)

    def test_random_state(self):
        random_state = np.random.RandomState(1)
        self.assertEqual(type(check_random_state(random_state)), np.random.RandomState)

    def test_error_random_state(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            check_random_state('junk')

    def test_check_seeds(self):
        n = 10
        seeds_array = -np.ones(n)
        seeds_array[:2] = np.arange(2)
        seeds_dict = {0: 0, 1: 1}
        labels_array = check_seeds(seeds_array, n)
        labels_dict = check_seeds(seeds_dict, n)

        self.assertTrue(np.allclose(labels_array, labels_dict))
        with self.assertRaises(ValueError):
            check_seeds(labels_array, 5)
        self.assertRaises(TypeError, check_seeds, 'toto', 3)
        with self.assertWarns(Warning):
            seeds_dict[0] = -1
            check_seeds(seeds_dict, n)

    def test_check_labels(self):
        with self.assertRaises(ValueError):
            check_labels(np.ones(3))
        labels = np.ones(5)
        labels[0] = 0
        classes, n_classes = check_labels(labels)
        self.assertTrue(np.equal(classes, np.arange(2)).all())
        self.assertEqual(n_classes, 2)

    def test_check_n_jobs(self):
        self.assertEqual(check_n_jobs(None), 1)
        self.assertEqual(check_n_jobs(-1), None)
        self.assertEqual(check_n_jobs(8), 8)

    def test_check_n_neighbors(self):
        with self.assertWarns(Warning):
            check_n_neighbors(10, 5)

    def test_adj_vector(self):
        n = 10
        vector1 = np.random.rand(n)
        vector2 = sparse.csr_matrix(vector1)
        adj1 = check_adjacency_vector(vector1)
        adj2 = check_adjacency_vector(vector2)

        self.assertEqual((adj1 - adj2).nnz, 0)
        self.assertEqual(adj1.shape, (1, n))

        with self.assertRaises(ValueError):
            check_adjacency_vector(vector1, 2 * n)

    def test_check_n_clusters(self):
        with self.assertRaises(ValueError):
            check_n_clusters(3, 2)
        with self.assertRaises(ValueError):
            check_n_clusters(0, 2, 1)

    def test_min_size(self):
        with self.assertRaises(ValueError):
            check_min_size(1, 3)

    def test_min_nnz(self):
        with self.assertRaises(ValueError):
            check_min_nnz(1, 3)

    def test_dendrogram(self):
        with self.assertRaises(ValueError):
            check_dendrogram(np.ones((3, 3)))

    def test_n_components(self):
        self.assertEqual(5, check_n_components(5, 10))
        with self.assertWarns(Warning):
            self.assertEqual(2, check_n_components(5, 2))

    def test_scaling(self):
        adjacency = cyclic_digraph(3)
        with self.assertRaises(ValueError):
            check_scaling(-1, adjacency, regularize=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for simplex.py"""

import unittest

import numpy as np
from scipy import sparse

from sknetwork.utils.check import is_proba_array
from sknetwork.utils.simplex import projection_simplex


class TestProjSimplex(unittest.TestCase):

    def test_array(self):
        x = np.random.rand(5)
        proj = projection_simplex(x)
        self.assertTrue(is_proba_array(proj))

        x = np.random.rand(4, 3)
        proj = projection_simplex(x)
        self.assertTrue(is_proba_array(proj))

    def test_csr(self):
        x = sparse.csr_matrix(np.ones((3, 3)))
        proj1 = projection_simplex(x)
        proj2 = projection_simplex(x.astype(bool))
        self.assertEqual(0, (proj1-proj2).nnz)

    def test_other(self):
        with self.assertRaises(TypeError):
            projection_simplex('toto')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for svd"""

import unittest

import numpy as np

from sknetwork.data import star_wars
from sknetwork.embedding import GSVD, PCA


class TestSVD(unittest.TestCase):

    def test_options(self):
        biadjacency = star_wars(metadata=False)
        n_row, n_col = biadjacency.shape
        min_dim = min(n_row, n_col) - 1
        gsvd = GSVD(n_components=5, regularization=0., solver='halko', relative_regularization=True)

        with self.assertWarns(Warning):
            gsvd.fit(biadjacency)
        self.assertEqual(gsvd.embedding_row_.shape, (n_row, min_dim))
        self.assertEqual(gsvd.embedding_col_.shape, (n_col, min_dim))

        gsvd = GSVD(n_components=1, regularization=0.1, solver='halko', relative_regularization=True)
        gsvd.fit(biadjacency)
        gsvd.predict(np.random.rand(n_col))

        pca = PCA(n_components=min_dim, solver='halko')
        pca.fit(biadjacency)
        self.assertEqual(pca.embedding_row_.shape, (n_row, min_dim))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.clustering.base import BaseClustering, BaseBiClustering
from sknetwork.clustering.postprocess import reindex_labels
from sknetwork.embedding.base import BaseEmbedding, BaseBiEmbedding
from sknetwork.embedding.svd import GSVD
from sknetwork.linalg.normalization import normalize
from sknetwork.utils.check import check_n_clusters
from sknetwork.utils.kmeans import KMeansDense
from sknetwork.utils.membership import membership_matrix


class KMeans(BaseClustering):
    """K-means applied in the embedding space.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_clusters :
        Number of desired clusters.
    embedding_method :
        Embedding method (default = GSVD in dimension 10, projected on the unit sphere).
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
            If ``True``, return the adjacency matrix of the graph between clusters.
    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix.
    adjacency_ : sparse.csr_matrix
        Adjacency matrix between clusters.

    Example
    -------
    >>> from sknetwork.clustering import KMeans
    >>> from sknetwork.data import karate_club
    >>> kmeans = KMeans(n_clusters=3)
    >>> adjacency = karate_club()
    >>> labels = kmeans.fit_transform(adjacency)
    >>> len(set(labels))
    3
    """
    def __init__(self, n_clusters: int = 8, embedding_method: BaseEmbedding = GSVD(10), sort_clusters: bool = True,
                 return_membership: bool = True, return_aggregate: bool = True):
        super(KMeans, self).__init__(sort_clusters=sort_clusters, return_membership=return_membership,
                                     return_aggregate=return_aggregate)
        self.n_clusters = n_clusters
        self.embedding_method = embedding_method

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'KMeans':
        """Apply embedding method followed by K-means.

        Parameters
        ----------
        adjacency:
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`KMeans`
        """
        n = adjacency.shape[0]
        check_n_clusters(self.n_clusters, n)

        embedding = self.embedding_method.fit_transform(adjacency)
        kmeans = KMeansDense(self.n_clusters)
        kmeans.fit(embedding)

        if self.sort_clusters:
            labels = reindex_labels(kmeans.labels_)
        else:
            labels = kmeans.labels_

        self.labels_ = labels
        self._secondary_outputs(adjacency)

        return self


class BiKMeans(KMeans, BaseBiClustering):
    """KMeans clustering of bipartite graphs applied in the embedding space.

    * Bigraphs

    Parameters
    ----------
    n_clusters :
        Number of clusters.
    embedding_method :
        Embedding method (default = GSVD in dimension 10, projected on the unit sphere).
    co_cluster :
        If ``True``, co-cluster rows and columns (default = ``False``).
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
            If ``True``, return the biadjacency matrix of the graph between clusters.
    Attributes
    ----------
    labels_ : np.ndarray
        Labels of the rows.
    labels_row_ : np.ndarray
        Labels of the rows (copy of **labels_**).
    labels_col_ : np.ndarray
        Labels of the columns. Only valid if **co_cluster** = `True`.
    membership_ : sparse.csr_matrix
        Membership matrix of the rows.
    membership_row_ : sparse.csr_matrix
        Membership matrix of the rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of the columns. Only valid if **co_cluster** = `True`.
    biadjacency_ : sparse.csr_matrix
        Biadjacency matrix of the graph between clusters.

    Example
    -------
    >>> from sknetwork.clustering import BiKMeans
    >>> from sknetwork.data import movie_actor
    >>> bikmeans = BiKMeans()
    >>> biadjacency = movie_actor()
    >>> labels = bikmeans.fit_transform(biadjacency)
    >>> len(labels)
    15
    """
    def __init__(self, n_clusters: int = 2, embedding_method: BaseBiEmbedding = GSVD(10), co_cluster: bool = False,
                 sort_clusters: bool = True, return_membership: bool = True, return_aggregate: bool = True):
        super(BiKMeans, self).__init__(sort_clusters=sort_clusters, return_membership=return_membership,
                                       return_aggregate=return_aggregate, n_clusters=n_clusters,
                                       embedding_method=embedding_method)
        self.co_cluster = co_cluster

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiKMeans':
        """Apply embedding method followed by clustering to the graph.

        Parameters
        ----------
        biadjacency:
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiKMeans`
        """
        n_row, n_col = biadjacency.shape
        check_n_clusters(self.n_clusters, n_row)

        method = self.embedding_method
        method.fit(biadjacency)

        if self.co_cluster:
            embedding = np.vstack((method.embedding_row_, method.embedding_col_))
        else:
            embedding = method.embedding_

        kmeans = KMeansDense(self.n_clusters)
        kmeans.fit(embedding)

        if self.sort_clusters:
            labels = reindex_labels(kmeans.labels_)
        else:
            labels = kmeans.labels_

        self.labels_ = labels
        if self.co_cluster:
            self._split_vars(n_row)
        else:
            self.labels_row_ = labels

        if self.return_membership:
            membership_row = membership_matrix(self.labels_row_, n_labels=self.n_clusters)
            if self.labels_col_ is not None:
                membership_col = membership_matrix(self.labels_col_, n_labels=self.n_clusters)
                self.membership_row_ = normalize(biadjacency.dot(membership_col))
                self.membership_col_ = normalize(biadjacency.T.dot(membership_row))
            else:
                self.membership_row_ = normalize(biadjacency.dot(biadjacency.T.dot(membership_row)))
            self.membership_ = self.membership_row_

        if self.return_aggregate:
            membership_row = membership_matrix(self.labels_row_, n_labels=self.n_clusters)
            biadjacency_ = sparse.csr_matrix(membership_row.T.dot(biadjacency))
            if self.labels_col_ is not None:
                membership_col = membership_matrix(self.labels_col_, n_labels=self.n_clusters)
                biadjacency_ = biadjacency_.dot(membership_col)
            self.biadjacency_ = biadjacency_

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for Louvain embedding"""
import unittest

import numpy as np

from sknetwork.data.test_graphs import test_graph, test_bigraph
from sknetwork.embedding import LouvainEmbedding, BiLouvainEmbedding


class TestLouvainEmbedding(unittest.TestCase):

    def test_predict(self):
        louvain = LouvainEmbedding()
        louvain.fit(test_graph())
        embedding_vector = louvain.predict(np.array([1, 0, 0, 0, 1, 1, 0, 0, 0, 1]))
        self.assertEqual(embedding_vector.shape[0], 1)

        bilouvain = BiLouvainEmbedding()
        bilouvain.fit(test_bigraph())
        embedding_vector = bilouvain.predict(np.array([1, 0, 0, 0, 1, 1]))
        self.assertEqual(embedding_vector.shape[0], 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for clustering post-processing"""

import unittest

import numpy as np

from sknetwork.clustering import reindex_labels


class TestClusteringPostProcessing(unittest.TestCase):

    def test_reindex_clusters(self):
        truth = np.array([1, 1, 2, 0, 0, 0])

        labels = np.array([0, 0, 1, 2, 2, 2])
        output = reindex_labels(labels)
        self.assertTrue(np.array_equal(truth, output))

        labels = np.array([0, 0, 5, 2, 2, 2])
        output = reindex_labels(labels, consecutive=False)
        self.assertTrue(np.array_equal(truth, output))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for pagerank.py"""

import unittest

import numpy as np

from sknetwork.data.models import cyclic_digraph
from sknetwork.data.test_graphs import test_bigraph
from sknetwork.ranking.pagerank import PageRank, BiPageRank
from sknetwork.utils import co_neighbor_graph


class TestPageRank(unittest.TestCase):

    def setUp(self) -> None:
        """Cycle graph for tests."""
        self.n = 5
        self.adjacency = cyclic_digraph(self.n)
        self.truth = np.ones(self.n) / self.n

    def test_params(self):
        with self.assertRaises(ValueError):
            PageRank(damping_factor=1789)

    def test_solvers(self):
        for solver in ['piteration', 'lanczos', 'bicgstab', 'RH']:
            pr = PageRank(solver=solver)
            scores = pr.fit_transform(self.adjacency)
            self.assertAlmostEqual(0, np.linalg.norm(scores - self.truth))

        PageRank(solver='diteration').fit_transform(self.adjacency)

        with self.assertRaises(ValueError):
            PageRank(solver='toto').fit_transform(self.adjacency)

    def test_seeding(self):
        pr = PageRank()
        seeds_array = np.zeros(self.n)
        seeds_array[0] = 1.
        seeds_dict = {0: 1}

        scores1 = pr.fit_transform(self.adjacency, seeds_array)
        scores2 = pr.fit_transform(self.adjacency, seeds_dict)
        self.assertAlmostEqual(np.linalg.norm(scores1 - scores2), 0.)

    def test_damping(self):
        pr = PageRank(damping_factor=0.99)
        scores = pr.fit_transform(self.adjacency)
        self.assertAlmostEqual(np.linalg.norm(scores - self.truth), 0.)

        pr = PageRank(damping_factor=0.01)
        scores = pr.fit_transform(self.adjacency)
        self.assertAlmostEqual(np.linalg.norm(scores - self.truth), 0.)

    def test_copagerank(self):
        seeds = {0: 1}
        biadjacency = test_bigraph()

        adjacency = co_neighbor_graph(biadjacency, method='exact', normalized=True)
        scores1 = BiPageRank(damping_factor=0.85, solver='lanczos').fit_transform(biadjacency, seeds)
        scores2 = PageRank(damping_factor=0.85**2, solver='lanczos').fit_transform(adjacency, seeds)
        self.assertAlmostEqual(np.linalg.norm(scores1 - scores2), 0., places=6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Jul 24, 2019
@author: Nathan de Lara <ndelara@enst.fr>
@author: Quentin Lutz <qlutz@enst.fr>
"""
from typing import Tuple, Optional, Union

import numpy as np
from scipy import sparse

from sknetwork.utils.check import is_symmetric, is_square, check_format


def connected_components(adjacency: sparse.csr_matrix, connection: str = 'weak') -> np.ndarray:
    """Extract the connected components of the graph.

    * Graphs
    * Digraphs

    Based on SciPy (scipy.sparse.csgraph.connected_components).

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    connection :
        Must be ``'weak'`` (default) or ``'strong'``. The type of connection to use for directed graphs.

    Returns
    -------
    labels : np.ndarray
        Connected component of each node.
    """
    adjacency = check_format(adjacency)
    if len(adjacency.data) == 0:
        raise ValueError('The graph is empty (no edge).')
    return sparse.csgraph.connected_components(adjacency, not is_symmetric(adjacency), connection, True)[1]


def largest_connected_component(adjacency: Union[sparse.csr_matrix, np.ndarray], return_labels: bool = False):
    """Extract the largest connected component of a graph. Bipartite graphs are treated as undirected.

    * Graphs
    * Digraphs
    * Bigraphs

    Parameters
    ----------
    adjacency :
        Adjacency or biadjacency matrix of the graph.
    return_labels : bool
        Whether to return the indices of the new nodes in the original graph.

    Returns
    -------
    new_adjacency : sparse.csr_matrix
        Adjacency or biadjacency matrix of the largest connected component.
    indices : array or tuple of array
        Indices of the nodes in the original graph. For biadjacency matrices,
        ``indices[0]`` corresponds to the rows and ``indices[1]`` to the columns.
    """
    adjacency = check_format(adjacency)
    n_row, n_col = adjacency.shape
    if not is_square(adjacency):
        bipartite: bool = True
        full_adjacency = sparse.bmat([[None, adjacency], [adjacency.T, None]], format='csr')
    else:
        bipartite: bool = False
        full_adjacency = adjacency

    labels = connected_components(full_adjacency)
    unique_labels, counts = np.unique(labels, return_counts=True)
    component_label = unique_labels[np.argmax(counts)]
    component_indices = np.where(labels == component_label)[0]

    if bipartite:
        split_ix = np.searchsorted(component_indices, n_row)
        row_ix, col_ix = component_indices[:split_ix], component_indices[split_ix:] - n_row
    else:
        row_ix, col_ix = component_indices, component_indices
    new_adjacency = adjacency[row_ix, :]
    new_adjacency = (new_adjacency.tocsc()[:, col_ix]).tocsr()

    if return_labels:
        if bipartite:
            return new_adjacency, (row_ix, col_ix)
        else:
            return new_adjacency, row_ix
    else:
        return new_adjacency


def is_bipartite(adjacency: sparse.csr_matrix, return_biadjacency: bool = False) \
        -> Union[bool, Tuple[bool, Optional[sparse.csr_matrix], Optional[np.ndarray], Optional[np.ndarray]]]:
    """Check whether an undirected graph is bipartite.

    * Graphs

    Parameters
    ----------
    adjacency :
       Adjacency matrix of the graph (symmetric).
    return_biadjacency :
        If ``True``, return a biadjacency matrix of the graph if bipartite.

    Returns
    -------
    is_bipartite : bool
        A boolean denoting if the graph is bipartite.
    biadjacency : sparse.csr_matrix
        A biadjacency matrix of the graph if bipartite (optional).
    rows : np.ndarray
        Index of rows in the original graph (optional).
    cols : np.ndarray
        Index of columns in the original graph (optional).
    """
    if not is_symmetric(adjacency):
        raise ValueError('The graph must be undirected.')
    if adjacency.diagonal().any():
        if return_biadjacency:
            return False, None, None, None
        else:
            return False
    n = adjacency.indptr.shape[0] - 1
    coloring = np.full(n, -1, dtype=int)
    exists_remaining = n
    while exists_remaining:
        src = np.argwhere(coloring == -1)[0, 0]
        next_nodes = [src]
        coloring[src] = 0
        exists_remaining -= 1
        while next_nodes:
            node = next_nodes.pop()
            for neighbor in adjacency.indices[adjacency.indptr[node]:adjacency.indptr[node + 1]]:
                if coloring[neighbor] == -1:
                    coloring[neighbor] = 1 - coloring[node]
                    next_nodes.append(neighbor)
                    exists_remaining -= 1
                elif coloring[neighbor] == coloring[node]:
                    if return_biadjacency:
                        return False, None, None, None
                    else:
                        return False
    if return_biadjacency:
        rows = np.argwhere(coloring == 0).ravel()
        cols = np.argwhere(coloring == 1).ravel()
        return True, adjacency[rows, :][:, cols], rows, cols
    else:
        return True


def is_acyclic(adjacency: sparse.csr_matrix) -> bool:
    """Check whether a graph has no cycle.

    Parameters
    ----------
    adjacency:
        Adjacency matrix of the graph.

    Returns
    -------
    is_acyclic : bool
        A boolean with value True if the graph has no cycle and False otherwise
    """
    n_nodes = adjacency.shape[0]
    n_cc = sparse.csgraph.connected_components(adjacency, (not is_symmetric(adjacency)), 'strong', False)
    if n_cc == n_nodes:
        # check for self-loops has they always induce a cycle
        return (adjacency.diagonal() == 0).all()
    else:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for DiffusionClassifier"""

import unittest

from sknetwork.classification import DiffusionClassifier
from sknetwork.data.test_graphs import *


class TestDiffusionClassifier(unittest.TestCase):

    def test_parallel(self):
        adjacency = test_graph()
        seeds = {0: 0, 1: 1}

        clf1 = DiffusionClassifier(n_jobs=None)
        clf2 = DiffusionClassifier(n_jobs=-1)

        labels1 = clf1.fit_transform(adjacency, seeds)
        labels2 = clf2.fit_transform(adjacency, seeds)

        self.assertTrue(np.allclose(labels1, labels2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import unittest

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import norm

from sknetwork.utils import KNNDense, CNNDense


class TestKNN(unittest.TestCase):

    def test_basics(self):
        x = np.array([[-2, -1], [-2, 1], [2, 1], [2, -1]])
        knn = KNNDense(n_neighbors=1)
        knn.fit(x)
        truth = np.zeros(16).reshape((4, 4))
        truth[0, 1] = 1
        truth[2, 3] = 1
        truth = sparse.csr_matrix(truth + truth.T)
        self.assertAlmostEqual(norm(truth - knn.adjacency_), 0)

    def test_pknn(self):
        x = np.array([(0, 4), (1, 0), (2, 1), (3, 2), (4, 3)])
        pknn = CNNDense(n_neighbors=1)
        adj = pknn.fit_transform(x)
        self.assertEqual(adj.shape, (5, 5))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 2019
@author: Thomas Bonald <bonald@enst.fr>
@author: Bertrand Charpentier <bertrand.charpentier@live.fr>
@author: Quentin Lutz <qlutz@enst.fr>
"""

import copy
from collections import defaultdict
from typing import Optional, Union, Tuple

import numpy as np

from sknetwork.utils.check import check_n_clusters, check_dendrogram


def reorder_dendrogram(dendrogram: np.ndarray):
    """Reorder the dendrogram in non-decreasing order of height."""
    n = dendrogram.shape[0] + 1
    order = np.zeros((2, n - 1), float)
    order[0] = np.max(dendrogram[:, :2], axis=1)
    order[1] = dendrogram[:, 2]
    index = np.lexsort(order)
    dendrogram_new = dendrogram[index]
    index_new = np.arange(2 * n - 1)
    index_new[n + index] = np.arange(n, 2 * n - 1)
    dendrogram_new[:, 0] = index_new[dendrogram_new[:, 0].astype(int)]
    dendrogram_new[:, 1] = index_new[dendrogram_new[:, 1].astype(int)]
    return dendrogram_new


def get_labels(dendrogram: np.ndarray, cluster: dict, sort_clusters: bool, return_dendrogram: bool):
    """Returns the labels from clusters."""
    n = dendrogram.shape[0] + 1
    n_clusters = len(cluster)
    clusters = list(cluster.values())
    index = None
    if sort_clusters:
        sizes = np.array([len(nodes) for nodes in clusters])
        index = np.argsort(-sizes)
        clusters = [clusters[i] for i in index]

    labels = np.zeros(n, dtype=int)
    for label, nodes in enumerate(clusters):
        labels[nodes] = label

    if return_dendrogram:
        indices_clusters = np.array(list(cluster.keys()))
        if sort_clusters:
            indices_clusters = indices_clusters[index]
        index_new = np.zeros(2 * n - 1, int)
        index_new[np.array(indices_clusters)] = np.arange(n_clusters)
        index_new[- n_clusters + 1:] = np.arange(n_clusters, 2 * n_clusters - 1)
        dendrogram_new = dendrogram[- n_clusters + 1:].copy()
        dendrogram_new[:, 0] = index_new[dendrogram_new[:, 0].astype(int)]
        dendrogram_new[:, 1] = index_new[dendrogram_new[:, 1].astype(int)]
        return labels, dendrogram_new
    else:
        return labels


def cut_straight(dendrogram: np.ndarray, n_clusters: Optional[int] = None, threshold: Optional[float] = None,
                 sort_clusters: bool = True, return_dendrogram: bool = False) \
                -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Cut a dendrogram and return the corresponding clustering.

    Parameters
    ----------
    dendrogram:
        Dendrogram.
    n_clusters :
        Number of clusters (optional).
        The number of clusters can be larger than n_clusters in case of equal heights in the dendrogram.
    threshold :
        Threshold on height (optional).
        If both n_clusters and threshold are ``None``, n_clusters is set to 2.
    sort_clusters :
        If ``True``,  sorts clusters in decreasing order of size.
    return_dendrogram :
        If ``True``, returns the dendrogram formed by the clusters up to the root.
    Returns
    -------
    labels : np.ndarray
        Cluster of each node.
    dendrogram_aggregate : np.ndarray
        Dendrogram starting from clusters (leaves = clusters).

    Example
    -------
    >>> from sknetwork.hierarchy import cut_straight
    >>> dendrogram = np.array([[0, 1, 0, 2], [2, 3, 1, 3]])
    >>> cut_straight(dendrogram)
    array([0, 0, 1])
    """
    check_dendrogram(dendrogram)
    n = dendrogram.shape[0] + 1

    if return_dendrogram and not np.all(np.diff(dendrogram[:, 2]) >= 0):
        raise ValueError("The third column of the dendrogram must be non-decreasing.")

    cluster = {i: [i] for i in range(n)}
    if n_clusters is None:
        if threshold is None:
            n_clusters = 2
        else:
            n_clusters = n
    else:
        check_n_clusters(n_clusters, n, n_min=1)
    cut = np.sort(dendrogram[:, 2])[n - n_clusters]
    if threshold is not None:
        cut = max(cut, threshold)
    for t in range(n - 1):
        i = int(dendrogram[t][0])
        j = int(dendrogram[t][1])
        if dendrogram[t][2] < cut and i in cluster and j in cluster:
            cluster[n + t] = cluster.pop(i) + cluster.pop(j)

    return get_labels(dendrogram, cluster, sort_clusters, return_dendrogram)


def cut_balanced(dendrogram: np.ndarray, max_cluster_size: int = 20, sort_clusters: bool = True,
                 return_dendrogram: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Cuts a dendrogram with a constraint on the cluster size and returns the corresponding clustering.

    Parameters
    ----------
    dendrogram:
        Dendrogram
    max_cluster_size :
        Maximum size of each cluster.
    sort_clusters :
        If ``True``, sort labels in decreasing order of cluster size.
    return_dendrogram :
        If ``True``, returns the dendrogram formed by the clusters up to the root.
    Returns
    -------
    labels : np.ndarray
        Label of each node.
    dendrogram_aggregate : np.ndarray
        Dendrogram starting from clusters (leaves = clusters).

    Example
    -------
    >>> from sknetwork.hierarchy import cut_balanced
    >>> dendrogram = np.array([[0, 1, 0, 2], [2, 3, 1, 3]])
    >>> cut_balanced(dendrogram, 2)
    array([0, 0, 1])
    """
    check_dendrogram(dendrogram)
    n = dendrogram.shape[0] + 1
    if max_cluster_size < 2 or max_cluster_size > n:
        raise ValueError("The maximum cluster size must be between 2 and the number of nodes.")

    cluster = {i: [i] for i in range(n)}
    for t in range(n - 1):
        i = int(dendrogram[t][0])
        j = int(dendrogram[t][1])
        if i in cluster and j in cluster and len(cluster[i]) + len(cluster[j]) <= max_cluster_size:
            cluster[n + t] = cluster.pop(i) + cluster.pop(j)

    return get_labels(dendrogram, cluster, sort_clusters, return_dendrogram)


def aggregate_dendrogram(dendrogram: np.ndarray, n_clusters: int = 2, return_counts: bool = False) \
                        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Aggregate a dendrogram in order to get a certain number of leaves.
    The leaves in the output dendrogram correspond to subtrees in the input one.

    Parameters
    ----------
    dendrogram:
        The input to aggregate.
    n_clusters:
        Number of clusters (or leaves) to keep.
    return_counts
        If ``True``, returns an array of counts corresponding to the sizes of the merged subtrees.
        The sum of the counts is equal to the number of samples in the input dendrogram.

    Returns
    -------
    new_dendrogram:
        Aggregated dendrogram. The nodes are reindexed from 0.
    counts:
        Size of the subtrees corresponding to each leaf in new_dendrogram.
    """
    n_nodes: int = dendrogram.shape[0] + 1
    check_n_clusters(n_clusters, n_nodes, n_min=1)

    new_dendrogram = dendrogram[n_nodes - n_clusters:].copy()
    node_indices = np.array(sorted(set(new_dendrogram[:, 0]).union(set(new_dendrogram[:, 1]))))
    new_index = {ix: i for i, ix in enumerate(node_indices)}

    for j in range(2):
        for i in range(new_dendrogram.shape[0]):
            new_dendrogram[i, j] = new_index[new_dendrogram[i, j]]

    if return_counts:
        leaves = node_indices[:n_clusters].astype(int)
        leaves_indices = leaves - n_nodes
        counts = dendrogram[leaves_indices, 3]

        return new_dendrogram, counts.astype(int)
    else:
        return new_dendrogram


def get_index(tree):
    """Reindex a dendrogram from the leaves

    Parameters
    ----------
    tree:
        The tree to be indexed

    Returns
    -------
    index:
        The index of the root of the given tree
    """
    if type(tree) != list:
        return tree
    else:
        return np.max([get_index(t) for t in tree])


def get_dendrogram(tree, dendrogram=None, index=None, depth=0, size=None, copy_tree=False):
    """Get dendrogram from tree.

    Parameters
    ----------
    tree :
        The initial tree
    dendrogram :
        Intermediary dendrogram for recursive use
    index :
        Intermediary index for recursive use
    depth :
        Current depth for recursive use
    size :
        Current leaf count for recursive use
    copy_tree :
        If ``True``, ensure the passed tree remains unchanged.

    Returns
    -------
    dendrogram`:
        The reordered dendrogram
    index :
        The indexing array
    """
    if copy_tree:
        return get_dendrogram(copy.deepcopy(tree))
    else:
        if dendrogram is None:
            dendrogram = []
        if index is None:
            index = get_index(tree)
        if size is None:
            size = defaultdict(lambda: 1)
        if len(tree) > 1:
            lengths = np.array([len(t) for t in tree])
            if np.max(lengths) == 1:
                # merge all
                i = tree.pop()[0]
                j = tree.pop()[0]
                s = size[i] + size[j]
                dendrogram.append([i, j, float(-depth), s])
                index += 1
                while len(tree):
                    s += 1
                    dendrogram.append([index, tree.pop()[0], float(-depth), s])
                    index += 1
                size[index] = s
                tree.append(index)
                return dendrogram, index
            else:
                i = np.argwhere(lengths > 1).ravel()[0]
                dendrogram_, index_ = get_dendrogram(tree[i], None, index, depth + 1, size)
                dendrogram += dendrogram_
                return get_dendrogram(tree, dendrogram, index_, depth, size)
        else:
            return dendrogram, index


def split_dendrogram(dendrogram: np.ndarray, shape: tuple):
    """Split the dendrogram of a bipartite graph into 2 dendrograms, one for each part.

    Parameters
    ----------
    dendrogram :
        Dendrogram of the bipartite graph.
    shape :
        Shape of the biadjacency matrix.
    Returns
    -------
    dendrogram_row :
        Dendrogram for the rows.
    dendrogram_col :
        Dendrogram for the columns.
    """
    n1, n2 = shape
    dendrogram_row = []
    dendrogram_col = []
    id_row_new = n1
    id_col_new = n2
    size_row = {i: 1 for i in range(n1)}
    size_col = {i + n1: 1 for i in range(n2)}
    id_row = {i: i for i in range(n1)}
    id_col = {i + n1: i for i in range(n2)}

    for t in range(n1 + n2 - 1):
        i = dendrogram[t, 0]
        j = dendrogram[t, 1]

        if i in id_row and j in id_row:
            size_row[n1 + n2 + t] = size_row.pop(i) + size_row.pop(j)
            id_row[n1 + n2 + t] = id_row_new
            dendrogram_row.append([id_row.pop(i), id_row.pop(j), dendrogram[t, 2], size_row[n1 + n2 + t]])
            id_row_new += 1
        elif i in id_row:
            size_row[n1 + n2 + t] = size_row.pop(i)
            id_row[n1 + n2 + t] = id_row.pop(i)
        elif j in id_row:
            size_row[n1 + n2 + t] = size_row.pop(j)
            id_row[n1 + n2 + t] = id_row.pop(j)

        if i in id_col and j in id_col:
            size_col[n1 + n2 + t] = size_col.pop(i) + size_col.pop(j)
            id_col[n1 + n2 + t] = id_col_new
            dendrogram_col.append([id_col.pop(i), id_col.pop(j), dendrogram[t, 2], size_col[n1 + n2 + t]])
            id_col_new += 1
        elif i in id_col:
            size_col[n1 + n2 + t] = size_col.pop(i)
            id_col[n1 + n2 + t] = id_col.pop(i)
        elif j in id_col:
            size_col[n1 + n2 + t] = size_col.pop(j)
            id_col[n1 + n2 + t] = id_col.pop(j)

    return np.array(dendrogram_row), np.array(dendrogram_col)




#!/usr/bin/env python3
# coding: utf-8
"""
Created on Sep 2020
@author: Quentin Lutz <qlutz@enst.fr>
"""
from typing import Optional, Union

import numpy as np
from scipy import sparse

from sknetwork.clustering.louvain import BiLouvain
from sknetwork.embedding.base import BaseBiEmbedding, BaseEmbedding
from sknetwork.utils.check import check_random_state, check_adjacency_vector, check_nonnegative
from sknetwork.utils.membership import membership_matrix


class BiLouvainEmbedding(BaseBiEmbedding):
    """Embedding of bipartite graphs from a clustering obtained with Louvain.

    Parameters
    ----------
    resolution : float
        Resolution parameter.
    merge_isolated : bool
        Denotes if clusters consisting of just one node should be merged.
    modularity : str
        Which objective function to maximize. Can be ``'dugue'``, ``'newman'`` or ``'potts'``.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    random_state :
        Random number generator or random seed. If None, numpy.random is used.

    Attributes
    ----------
    embedding_ : array, shape = (n, n_components)
        Embedding of the nodes.
    embedding_row_ : array, shape = (n_row, n_components)
        Embedding of the rows (copy of **embedding_**).
    embedding_col_ : array, shape = (n_col, n_components)
        Embedding of the columns.

    Example
    -------
    >>> from sknetwork.embedding import BiLouvainEmbedding
    >>> from sknetwork.data import movie_actor
    >>> bilouvain = BiLouvainEmbedding()
    >>> biadjacency = movie_actor()
    >>> embedding = bilouvain.fit_transform(biadjacency)
    >>> embedding.shape
    (15, 5)
    """
    def __init__(self, resolution: float = 1, merge_isolated: bool = True, modularity: str = 'dugue',
                 tol_optimization: float = 1e-3, tol_aggregation: float = 1e-3, n_aggregations: int = -1,
                 shuffle_nodes: bool = False, random_state: Optional[Union[np.random.RandomState, int]] = None):
        super(BiLouvainEmbedding, self).__init__()
        self.resolution = np.float32(resolution)
        self.modularity = modularity.lower()
        self.tol_optimization = np.float32(tol_optimization)
        self.tol_aggregation = tol_aggregation
        self.n_aggregations = n_aggregations
        self.shuffle_nodes = shuffle_nodes
        self.random_state = check_random_state(random_state)
        self.merge_isolated = merge_isolated

        self.labels_ = None

    def fit(self, biadjacency: sparse.csr_matrix):
        """Embedding of bipartite graphs from a clustering obtained with Louvain.

        Parameters
        ----------
        biadjacency:
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiLouvainEmbedding`
        """
        bilouvain = BiLouvain(resolution=self.resolution, modularity=self.modularity,
                              tol_optimization=self.tol_optimization, tol_aggregation=self.tol_aggregation,
                              n_aggregations=self.n_aggregations, shuffle_nodes=self.shuffle_nodes, sort_clusters=True,
                              return_membership=True, return_aggregate=True, random_state=self.random_state)
        bilouvain.fit(biadjacency)

        self.labels_ = bilouvain.labels_

        embedding_row = bilouvain.membership_row_
        embedding_col = bilouvain.membership_col_

        if self.merge_isolated:
            _, counts_row = np.unique(bilouvain.labels_row_, return_counts=True)
            n_isolated_nodes_row = (counts_row == 1).sum()
            if n_isolated_nodes_row:
                size_row = (biadjacency.shape[0], len(counts_row))
                embedding_row.resize(size_row)
                labels_row = bilouvain.labels_row_
                labels_row[-n_isolated_nodes_row:] = labels_row[-n_isolated_nodes_row]
                merge_labels_row = np.arange(len(counts_row), dtype=int)
                merge_labels_row[-n_isolated_nodes_row:] = merge_labels_row[-n_isolated_nodes_row]
                combiner_row = membership_matrix(merge_labels_row)
                embedding_row = embedding_row.dot(combiner_row)
                self.labels_ = labels_row

            _, counts_col = np.unique(bilouvain.labels_col_, return_counts=True)
            n_isolated_nodes_col = (counts_col == 1).sum()
            if n_isolated_nodes_col:
                size_col = (biadjacency.shape[1], len(counts_col))
                embedding_col.resize(size_col)
                merge_labels_col = np.arange(embedding_col.shape[1], dtype=int)
                merge_labels_col[-n_isolated_nodes_col:] = merge_labels_col[-n_isolated_nodes_col]
                combiner_col = membership_matrix(merge_labels_col)
                embedding_col = embedding_col.dot(combiner_col)

        self.embedding_row_ = embedding_row
        self.embedding_col_ = embedding_col
        self.embedding_ = self.embedding_row_

        return self

    def predict(self, adjacency_vectors: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Predict the embedding of new rows, defined by their adjacency vectors.

        Parameters
        ----------
        adjacency_vectors :
            Adjacency vectors of rows.
            Array of shape (n_col,) (single vector) or (n_vectors, n_col)

        Returns
        -------
        embedding_vectors : np.ndarray
            Embedding of the nodes.
        """
        self._check_fitted()
        n = self.embedding_.shape[0]

        adjacency_vectors = check_adjacency_vector(adjacency_vectors, n)
        check_nonnegative(adjacency_vectors)
        membership = membership_matrix(self.labels_)

        return adjacency_vectors.dot(membership)


class LouvainEmbedding(BaseEmbedding):
    """Embedding of graphs from a clustering obtained with Louvain.

    Parameters
    ----------
    resolution :
        Resolution parameter.
    merge_isolated : bool
        Denotes if clusters consisting of just one node should be merged.
    modularity : str
        Which objective function to maximize. Can be ``'dugue'``, ``'newman'`` or ``'potts'``.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    random_state :
        Random number generator or random seed. If None, numpy.random is used.

    Attributes
    ----------
    embedding_ : array, shape = (n, n_components)
        Embedding of the nodes.

    Example
    -------
    >>> from sknetwork.embedding import LouvainEmbedding
    >>> from sknetwork.data import karate_club
    >>> louvain = LouvainEmbedding()
    >>> adjacency = karate_club()
    >>> embedding = louvain.fit_transform(adjacency)
    >>> embedding.shape
    (34, 7)
    """
    def __init__(self, resolution: float = 1, merge_isolated: bool = True, modularity: str = 'dugue',
                 tol_optimization: float = 1e-3, tol_aggregation: float = 1e-3, n_aggregations: int = -1,
                 shuffle_nodes: bool = False, random_state: Optional[Union[np.random.RandomState, int]] = None):
        super(LouvainEmbedding, self).__init__()
        self.resolution = np.float32(resolution)
        self.modularity = modularity.lower()
        self.tol_optimization = np.float32(tol_optimization)
        self.tol_aggregation = tol_aggregation
        self.n_aggregations = n_aggregations
        self.shuffle_nodes = shuffle_nodes
        self.random_state = check_random_state(random_state)
        self.merge_isolated = merge_isolated

        self.labels_ = None

    def fit(self, adjacency: sparse.csr_matrix):
        """Embedding of graphs from a clustering obtained with Louvain.

        Parameters
        ----------
        adjacency:
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`LouvainEmbedding`
        """
        bilouvain = BiLouvainEmbedding(resolution=self.resolution, merge_isolated=self.merge_isolated,
                                       modularity=self.modularity, tol_optimization=self.tol_optimization,
                                       tol_aggregation=self.tol_aggregation, n_aggregations=self.n_aggregations,
                                       shuffle_nodes=self.shuffle_nodes, random_state=self.random_state)
        bilouvain.fit(adjacency)
        self.labels_ = bilouvain.labels_
        self.embedding_ = bilouvain.embedding_

        return self

    def predict(self, adjacency_vectors: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Predict the embedding of new nodes, defined by their adjacency vectors.

        Parameters
        ----------
        adjacency_vectors :
            Adjacency vectors of nodes.
            Array of shape (n_col,) (single vector) or (n_vectors, n_col)

        Returns
        -------
        embedding_vectors : np.ndarray
            Embedding of the nodes.
        """
        self._check_fitted()
        n = self.embedding_.shape[0]

        adjacency_vectors = check_adjacency_vector(adjacency_vectors, n)
        check_nonnegative(adjacency_vectors)

        membership = membership_matrix(self.labels_)

        return adjacency_vectors.dot(membership)
